"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# pylint: disable=duplicate-code
from loom.eka import ChannelType, Circuit, PauliOperator, Stabilizer
from loom.eka.operations import Split
from loom.eka.utilities import Direction, Orientation
from loom.interpreter import InterpretationStep
from loom.interpreter.utilities import Cbit

from ..code_factory import RepetitionCode


# pylint: disable=too-many-locals
def split_consistency_check(
    interpretation_step: InterpretationStep, operation: Split
) -> RepetitionCode:
    """
    Check that the split operation can be performed on the given chain.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        InterpretationStep containing the block to split.
    operation : Split
        Split operation to perform.

    Returns
    -------
    RepetitionCode
        Block to split.

    Raises
    ------
    ValueError
        If the split position is larger than the length of the chain.
        If the split position is at the edge of the chain.
        If one of the chains has a length of one.
    """
    # Extract Block
    block = interpretation_step.get_block(operation.input_block_name)

    # Check valid Block type
    if not isinstance(block, RepetitionCode):
        raise ValueError(
            f"This split operation is not supported for {type(block)} blocks."
        )

    relative_split_position = operation.split_position
    left_boundary = block.boundary_qubits(Direction.LEFT)[0]
    right_boundary = block.boundary_qubits(Direction.RIGHT)[0]

    split_position = left_boundary + relative_split_position

    # Check that the split position is under correct parameters
    if split_position > right_boundary or split_position < left_boundary:
        raise ValueError(f"Split position {split_position} is outside the chain.")

    if split_position in [left_boundary, right_boundary]:
        raise ValueError("Split position cannot be at the edge of the chain.")

    if split_position in [left_boundary + 1, right_boundary - 1]:
        raise ValueError(
            "Split has to partition chain in units of at least two qubits."
        )

    # Check that the split operation can be performed
    if operation.orientation != Orientation.VERTICAL:
        raise ValueError("Only vertical splits are supported.")

    return block


def find_qubit_to_measure(
    block: RepetitionCode, split_position: int
) -> tuple[int, int]:
    """Find the qubit to measure in the split operation.

    Parameters
    ----------
    block : RepetitionCode
        Block to split.
    split_position : int
        Position where the chain will be split.

    Returns
    -------
    tuple[int, int]
        Qubit to be measured in the split operation.
    """
    boundary_qubit_position = block.boundary_qubits(Direction.LEFT)[0]
    qubit_to_measure = (boundary_qubit_position + split_position, 0)
    return qubit_to_measure


def create_split_circuit(
    interpretation_step: InterpretationStep,
    check_type: str,
    qubit_to_measure: tuple[int, int],
    circuit_name: str,
) -> tuple[Circuit, Cbit]:
    """Create the circuit for the split operation.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the chain to split.
    check_type : str
        Type of stabilizer to measure.
    qubits_to_measure : tuple[int, int]
        Qubit to be measured in the split operation.
    circuit_name : str
        Name of the circuit for the split operation.

    Returns
    -------
    tuple[Circuit, Cbit]
        Circuit for the split operation and the classical bit used in the circuit.
    """

    # Create classical channels for all data qubit measurements
    cbit = interpretation_step.get_new_cbit_MUT(f"c_{qubit_to_measure}")
    cbit_channel = interpretation_step.get_channel_MUT(
        f"{cbit[0]}_{cbit[1]}", channel_type=ChannelType.CLASSICAL
    )

    # Create a measurement circuit for every measured data qubit
    measure_circuit_seq = [
        Circuit(
            "Measurement",
            channels=[
                interpretation_step.get_channel_MUT(qubit_to_measure),
                cbit_channel,
            ],
        )
    ]

    # If needed, apply a basis change
    if check_type == "X":
        # If phaseflip code qubits can already be measured
        split_circuit_list = measure_circuit_seq
    else:
        # If bitflip code qubits need to be Hadamard transformed before measurement
        basis_change_circuit_seq = [
            Circuit(
                "H", channels=[interpretation_step.get_channel_MUT(qubit_to_measure)]
            )
        ]
        # Add basis change before measurement
        split_circuit_list = basis_change_circuit_seq + measure_circuit_seq

    # Build circuit
    split_circuit = Circuit(
        name=circuit_name,
        circuit=split_circuit_list,
    )
    return split_circuit, cbit


def find_new_stabilizers(
    block: RepetitionCode,
    qubit_to_measure: tuple[int, int],
) -> tuple[list[Stabilizer], list[Stabilizer]]:
    """Split the initial Block stabilizers into two lists of stabilizers, one for
    each Block.

    Parameters
    ----------
    block : RepetitionCode
        Initial block to be split
    qubits_to_measure : tuple[int,int]
        Qubit to be measured in the split operation.

    Returns
    -------
    tuple[list[Stabilizer], list[Stabilizer]]
        A tuple of two lists of stabilizers, one for each block.
    """

    # Find stabilizers which are completely removed
    stabs_to_remove = [
        stab
        for stab in block.stabilizers
        if any(q == qubit_to_measure for q in stab.data_qubits)
    ]

    # Create two sets of stabilizers, one for each new block
    new_stabilizers = [
        stab for stab in block.stabilizers if stab not in stabs_to_remove
    ]

    data_qubits_block_1 = set(
        q for q in block.data_qubits if q[0] < qubit_to_measure[0]
    )
    data_qubits_block_2 = set(
        q for q in block.data_qubits if q[0] > qubit_to_measure[0]
    )

    new_stabs_block_1 = [
        stab
        for stab in new_stabilizers
        if set(stab.data_qubits).issubset(data_qubits_block_1)
    ]
    new_stabs_block_2 = [
        stab
        for stab in new_stabilizers
        if set(stab.data_qubits).issubset(data_qubits_block_2)
    ]

    # Compute total set of remaining stabilizers
    remaining_stabs = (
        set(new_stabilizers) - set(new_stabs_block_1) - set(new_stabs_block_2)
    )

    # Consistency check that all new stabilizers are assigned to a block
    if len(remaining_stabs) != 0:
        raise ValueError(
            f"Stabilizers {remaining_stabs} are not assigned to any block."
        )

    return (
        new_stabs_block_1,
        new_stabs_block_2,
    )


def get_logical_operator_and_updates(
    int_step: InterpretationStep,
    block: RepetitionCode,
    check_type: str,
    qubit_to_measure: tuple[int, int],
    cbit: Cbit,
) -> tuple[
    list[list[PauliOperator]],
    list[list[PauliOperator]],
    list[dict[str, tuple[str, ...]]],
    list[dict[str, tuple[str, ...]]],
    list[dict[str, tuple[Cbit, ...]]],
    list[dict[str, tuple[Cbit, ...]]],
]:
    """Finds the logical operators for the new blocks after the split operation. The
    logical spanning the chain will be split in two, the logical at the left end remains
    there for the first chain and a new one is created for the second chain, on its left
    end.

    Parameters
    ----------
    int_step : InterpretationStep
        Interpretation step containing the block to split.
    block : RepetitionCode
        Block to split.
    check_type : str
        Type of stabilizer to measure.
    qubit_to_measure : tuple[int,int]
        Qubit to be measured in the split operation.
    cbit : Cbit
        Classical bit used to measure the qubit that splits the chain.

    Returns
    -------
    tuple[
        list[list[PauliOperator]],list[list[PauliOperator]],
        list[dict[str, tuple[str, ...]]],
        list[dict[str, tuple[str, ...]]],
        list[dict[str, tuple[Cbit, ...]]],
        list[dict[str, tuple[Cbit, ...]]],
    ]
        Tuple containing the new logical operators for the blocks, the evolution,
        and the updates.
    """

    other_check_type = "X" if check_type == "Z" else "Z"

    # block 1 is the left one
    qubits_block_1 = [q for q in block.data_qubits if q[0] < qubit_to_measure[0]]
    left_boundary_qubit_1 = min(qubits_block_1, key=lambda x: x[0])

    # block 2 is the right one
    qubits_block_2 = [q for q in block.data_qubits if q[0] > qubit_to_measure[0]]
    left_boundary_qubit_2 = min(qubits_block_2, key=lambda x: x[0])

    # Extract old long and short operators
    old_short_log, old_long_log = sorted(
        (block.logical_x_operators[0], block.logical_z_operators[0]),
        key=lambda x: len(x.data_qubits),
    )

    # Define new logical operators
    new_long_log_1 = PauliOperator(
        pauli=other_check_type * len(qubits_block_1), data_qubits=qubits_block_1
    )
    long_log_evolution_1 = {new_long_log_1.uuid: (old_long_log.uuid,)}
    long_log_updates_1 = {
        new_long_log_1.uuid: (cbit,)
    }  # Give the correction to the first operator

    new_long_log_2 = PauliOperator(
        pauli=other_check_type * len(qubits_block_2), data_qubits=qubits_block_2
    )
    long_log_evolution_2 = {new_long_log_2.uuid: (old_long_log.uuid,)}
    long_log_updates_2 = {}

    # Short logicals get placed in the left boundary in the new blocks
    new_short_log_1, required_stabs_1 = block.get_shifted_equivalent_logical_operator(
        left_boundary_qubit_1
    )
    id_required_stabs_1 = [stab.uuid for stab in required_stabs_1]

    new_short_log_2, required_stabs_2 = block.get_shifted_equivalent_logical_operator(
        left_boundary_qubit_2
    )
    id_required_stabs_2 = [stab.uuid for stab in required_stabs_2]

    # If logical was already in the default position, no evolution
    if id_required_stabs_1 == []:
        new_short_log_1 = old_short_log
        short_log_evolution_1 = {}
        short_log_update_1 = {}
    # If logical in left Block was shifted, add evolution
    else:
        short_log_evolution_1 = {
            new_short_log_1.uuid: tuple([old_short_log.uuid] + id_required_stabs_1)
        }
        short_log_cbits_1 = int_step.retrieve_cbits_from_stabilizers(
            required_stabs_1, block
        )
        short_log_update_1 = {new_short_log_1.uuid: short_log_cbits_1}

    # If logical was already in the default position, no evolution
    if id_required_stabs_2 == []:
        new_short_log_2 = old_short_log
        short_log_evolution_2 = {}
        short_log_update_2 = {}
    # If logical in right Block was shifted, add evolution
    else:
        short_log_evolution_2 = {
            new_short_log_2.uuid: tuple([old_short_log.uuid] + id_required_stabs_2)
        }
        short_log_cbits_2 = int_step.retrieve_cbits_from_stabilizers(
            required_stabs_2, block
        )
        short_log_update_2 = {new_short_log_2.uuid: short_log_cbits_2}

    # New logicals
    new_logs_1 = [[new_long_log_1], [new_short_log_1]]
    new_logs_2 = [[new_long_log_2], [new_short_log_2]]

    # New operator evolution and updates
    # Find cbits for the logical updates
    log_updates_1 = [long_log_updates_1, short_log_update_1]
    log_evolution_1 = [long_log_evolution_1, short_log_evolution_1]

    log_updates_2 = [long_log_updates_2, short_log_update_2]
    log_evolution_2 = [long_log_evolution_2, short_log_evolution_2]

    return (
        new_logs_1,
        new_logs_2,
        log_evolution_1,
        log_evolution_2,
        log_updates_1,
        log_updates_2,
    )


def split(
    interpretation_step: InterpretationStep,
    operation: Split,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Applicator to split a Repetition Code chain using the Split description.
    Since the Repetition code is a 1D chain, the split operation is always vertical.

    The algorithm is the following:

    - A.) DATA QUBITS

        - A.1) Find data qubit to be measured

    - B.) - CIRCUIT

        - B.1) Create a measurement circuit for the measured data qubit
        - B.2) Append the measurement circuits to the InterpretationStep circuit

    - C.) - STABILIZERS

        - C.1) Create two sets of stabilizers, one for each new block, after the split

    - D.) LOGICAL OPERATORS

        - D.1) Find logical operators which are not partially measured
        - D.2) Find logical operators which have to be split
        - D.3) Create new logical operators and final to initial mapping
        - D.4) Update `logical_x/z_evolution`
        - D.5) Update `logical_x/z_updates`

    - E.) NEW BLOCK AND NEW INTERPRETATION STEP

        - E.1) Create the new blocks
        - E.2) Update the block history

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the block to split.
    operation : Split
        Split operation description.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the
        previous operation.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Activating debug mode will enable commutation validation for Block.

    Returns
    -------
    InterpretationStep
        Interpretation step after the split operation.
    """

    # Check that the operation can be performed
    block = split_consistency_check(interpretation_step, operation)

    # Extract split position
    split_position = operation.split_position

    # Extract check type
    check_type = block.check_type

    # A) - DATA QUBITS
    #   A.1) Find data qubit to be measured
    qubit_to_measure = find_qubit_to_measure(block, split_position)

    # B) - CIRCUIT
    #    B.1) Create a measurement circuit for the measured data qubit
    circuit_name = f"split {block.unique_label} at {split_position}"
    split_circuit, cbit = create_split_circuit(
        interpretation_step, check_type, qubit_to_measure, circuit_name
    )

    #    B.2) Append the measurement circuits to the InterpretationStep circuit
    interpretation_step.append_circuit_MUT(split_circuit, same_timeslice)

    # C) - STABILIZERS
    #    C.1) Create two sets of stabilizers, one for each new block, after the split
    (
        new_stabs_block_1,
        new_stabs_block_2,
    ) = find_new_stabilizers(block, qubit_to_measure)

    # D) LOGICAL OPERATORS
    #    D.1) Extract logical operators and updates
    (
        new_logs_1,
        new_logs_2,
        log_evolution_1,
        log_evolution_2,
        log_updates_1,
        log_updates_2,
    ) = get_logical_operator_and_updates(
        interpretation_step, block, check_type, qubit_to_measure, cbit
    )

    # D.2) Update the logical operator history
    # Evolution and updates are returned in the order [long_log,short_log]
    # Depending on the check_type we invert the order of the evolution and
    # updates output, to match the order of variable assignment below
    ordering = 1 if check_type == "Z" else -1
    x_is_long_operator = check_type == "Z"

    new_log_x_ops_1, new_log_z_ops_1 = new_logs_1[::ordering]
    new_log_x_ops_2, new_log_z_ops_2 = new_logs_2[::ordering]
    logical_x_evolution_1, logical_z_evolution_1 = log_evolution_1[::ordering]
    logical_x_evolution_2, logical_z_evolution_2 = log_evolution_2[::ordering]
    logical_x_updates_1, logical_z_updates_1 = log_updates_1[::ordering]
    logical_x_updates_2, logical_z_updates_2 = log_updates_2[::ordering]

    interpretation_step.logical_x_evolution.update(
        logical_x_evolution_1 | logical_x_evolution_2
    )
    # Update the logical X operator updates:
    # The first block's operator always inherits updates, only the second block's short
    # operator also inherits updates
    for op_id, cbits in logical_x_updates_1.items():
        interpretation_step.update_logical_operator_updates_MUT("X", op_id, cbits, True)
    for op_id, cbits in logical_x_updates_2.items():
        interpretation_step.update_logical_operator_updates_MUT(
            "X", op_id, cbits, not x_is_long_operator
        )

    interpretation_step.logical_z_evolution.update(
        logical_z_evolution_1 | logical_z_evolution_2
    )
    # Update the logical Z operator updates:
    # The first block's operator always inherits updates, only the second block's short
    # operator also inherits updates
    for op_id, cbits in logical_z_updates_1.items():
        interpretation_step.update_logical_operator_updates_MUT("Z", op_id, cbits, True)
    for op_id, cbits in logical_z_updates_2.items():
        interpretation_step.update_logical_operator_updates_MUT(
            "Z", op_id, cbits, x_is_long_operator
        )

    # E) NEW BLOCK AND NEW INTERPRETATION STEP
    #    E.1) Create the new blocks
    new_block_1 = RepetitionCode(
        stabilizers=new_stabs_block_1,
        logical_x_operators=new_log_x_ops_1,
        logical_z_operators=new_log_z_ops_1,
        unique_label=operation.output_blocks_name[0],
        skip_validation=not debug_mode,
    )
    new_block_2 = RepetitionCode(
        stabilizers=new_stabs_block_2,
        logical_x_operators=new_log_x_ops_2,
        logical_z_operators=new_log_z_ops_2,
        unique_label=operation.output_blocks_name[1],
        skip_validation=not debug_mode,
    )
    #    E.2) Update the block history
    interpretation_step.update_block_history_and_evolution_MUT(
        new_blocks=(new_block_1, new_block_2), old_blocks=(block,)
    )

    return interpretation_step



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations
from functools import cached_property
from pydantic.dataclasses import dataclass
from ..utilities import dataclass_config


@dataclass(config=dataclass_config)
class LogicalMeasurement:
    """
    LogicalMeasurement acts as a wrapper to describe the type of logical measurement
    based on the block(s) involved and the pauli product of the observable.
    NOTE: Assumes that each block only consist of one logical qubit, and refers only to
    the latest logical measurement of the block(s).

    E.g. LogicalMeasurement(("block0", "block1"), "ZZ") describes the joint ZZ
    measurement between block0 and block1.

    Parameters
    ----------
    blocks : tuple[str, ...]
        List of names of blocks that are involved in the logical measurement.
    observable : str
        Pauli string describing the type of logical measurement.
    """

    blocks: tuple[str, ...]
    observable: str  # e.g. 'X', 'ZZ'

    @cached_property
    def _as_pairs(self) -> frozenset[tuple[str, str]]:
        """Internal representation as a frozenset of (block, observable) pairs.

        This ensures that two equivalent LogicalMeasurement objects
        (e.g., with the same blocks and observables in different orders)
        compare equal and can be used interchangeably in sets or dicts.
        """
        return frozenset(zip(self.blocks, self.observable, strict=True))

    def __eq__(self, other: LogicalMeasurement) -> bool:
        return self._as_pairs == other._as_pairs

    def __hash__(self) -> int:
        return hash(self._as_pairs)

    def __repr__(self) -> str:
        blocks_str = ", ".join(self.blocks)
        return (
            f"LogicalMeasurement(blocks=({blocks_str}), "
            f"observable='{self.observable}')"
        )



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# pylint: disable=line-too-long

from loom.eka import Circuit, Stabilizer, PauliOperator, SyndromeCircuit
from loom.interpreter import InterpretationStep

from ..operations import TransversalHadamard
from ..code_factory import RotatedSurfaceCode

# # pylint: disable=duplicate-code


def update_synd_circ(
    input_synd_circ: SyndromeCircuit, new_pauli: str
) -> SyndromeCircuit:
    """Updates the input Measurement SyndromeCircuit to measure a new pauli.
    The input SyndromeCircuit requires a specific design where the ancilla qubits
    are initialized and measured in the X basis with a CX, CY or CZ gate to the data
    qubit.
    The function replaces the two-qubit gate with a new gate that measures the new_pauli.
    If the new pauli is a Z, the function replaces the two-qubit gate with a CZ gate.
    If the new pauli is an X, the function replaces the two-qubit gate with a CX gate.
    If the new pauli is a Y, the function replaces the two-qubit gate with a CY gate.

    Parameters
    ----------
    input_synd_circ : SyndromeCircuit
        The input syndrome circuit.
    new_pauli : str
        The new pauli the circuit will be measuring now.

    Returns
    -------
        SyndromeCircuit: The updated syndrome circuit.
    """
    # Iterate through Syndrome Circuit and change the two-qubit gates.
    # Get actual circuit object.
    actual_circ_obj = input_synd_circ.circuit
    inner_circ_tuple = actual_circ_obj.circuit
    new_circ = tuple()

    # Check that Syndrome Circuit has the same Pauli Length as New Stabilizer
    if len(new_pauli) != len(input_synd_circ.pauli):
        raise ValueError(
            "The new stabilizer has a different pauli length than the input syndrome circuit."
        )

    iterator = 0
    for each_tick in inner_circ_tuple:
        # If tick is empty, add an empty tuple.
        if len(each_tick) == 0:
            new_circ += ((),)
            continue

        # Else, Iterate through each gate in the tick.
        for each_gate in each_tick:
            if each_gate.name in ["cz", "cx", "cy"]:
                # Recreate the gate with a new name.
                new_circ += (
                    (
                        Circuit(
                            name=f"c{new_pauli[iterator]}", channels=each_gate.channels
                        ),
                    ),
                )
                iterator += 1
            else:
                new_circ += ((each_gate,),)

    # Get new circuit name.
    new_name = (
        actual_circ_obj.name[: -1 * len(input_synd_circ.pauli)] + new_pauli.lower()
    )

    # Recreate Circuit Block with updated name.
    # Channels need to be assigned properly. Automatic assignment jumbles ordering.
    new_circ_obj = Circuit(
        name=new_name,
        circuit=new_circ,
        channels=actual_circ_obj.channels,
    )

    # Create new Syndrome Circuit object.
    new_synd_circ = SyndromeCircuit(
        pauli=new_pauli,
        name=new_name,
        circuit=new_circ_obj,
    )

    return new_synd_circ


def transversal_hadamard_syndrome_circuits(
    block: RotatedSurfaceCode,
    interpretation_step: InterpretationStep,
    new_stabs: list[Stabilizer],
    new_stabilizer_to_circuit: dict[str, str],
    new_synd_circs: tuple[SyndromeCircuit] = tuple(),
) -> tuple[tuple[SyndromeCircuit], dict[str, str]]:
    """For a given Block, InterpretationStep and list of new Stabilizers that
    result from a Transversal Hadamard operation, this generates the new SyndromeCircuit
    objects that are associated with the new Stabilizers. Along with the new
    SyndromeCircuit objects, it also generates a mapping of the new Stabilizers to
    the UUIDs of the new SyndromeCircuits. An initial collection of new SyndromeCircuits
    and a partially populated mapping of the new Stabilizers to the UUIDs of the new
    SyndromeCircuits can be provided as optional arguments.

    Parameters
    ----------
    block : RotatedSurfaceCode
        The block of the surface code.
    interpretation_step : InterpretationStep
        The interpretation step that contains the evolution of the new Stabilizers.
    new_stabs : list[Stabilizer]
        The new stabilizers that result from the Transversal Hadamard operation.
    new_stabilizer_to_circuit : dict[str, str]
        The partially populated mapping of the new Stabilizers to the UUIDs of the new
        SyndromeCircuits.
    new_synd_circs : tuple[SyndromeCircuit], optional
        The new SyndromeCircuit objects that are ALREADY associated with the new
        Stabilizers.

    Returns
    -------
    tuple[tuple[SyndromeCircuit], dict[str,str]]
        A tuple containing the new SyndromeCircuit objects and a dictionary mapping
        the new Stabilizers to the UUIDs of the new SyndromeCircuits.
    """
    # D) SYNDROMECIRCUITS
    # D.1) Update Syndrome Circuits with new UUIDs.
    for each_new_stab in new_stabs:
        # D.2) Determine the Syndrome Circuit(s) associated with the old Stabilizer(s).
        # If more than one stabilizers evolved into the same stabilizer, we only
        # consider the first one.
        old_stab_uuid = interpretation_step.stabilizer_evolution[each_new_stab.uuid][0]
        old_synd_circ_uuid = block.stabilizer_to_circuit[old_stab_uuid]
        selected_synd_circ = next(
            (
                each_selected_synd
                for each_selected_synd in block.syndrome_circuits
                if each_selected_synd.uuid == old_synd_circ_uuid
            )
        )
        # D.3) Update the Syndrome Circuit(s) for the new Stabilizer(s).
        updated_synd_circ = update_synd_circ(selected_synd_circ, each_new_stab.pauli)

        # D.4) Update Collection of New Syndrome Circuits and New Stabilizer to Circuit Mapping
        # We consider Syndrome Circuits duplicates if all their properties except UUID are the same.
        duplicate_synd_circ = next(
            (
                new_synd_circ
                for new_synd_circ in new_synd_circs
                if new_synd_circ == updated_synd_circ
            ),
            None,
        )
        if not duplicate_synd_circ:
            # If there is no duplicate, we add the new Syndrome Circuit to the collection.
            new_synd_circs += (updated_synd_circ,)
            # We map the stabilizer to the new Syndrome Circuit's UUID.
            new_stabilizer_to_circuit.update(
                {each_new_stab.uuid: updated_synd_circ.uuid}
            )
        else:
            # If there is a duplicate, we map the stabilizer to the duplicate's UUID instead.
            new_stabilizer_to_circuit.update(
                {each_new_stab.uuid: duplicate_synd_circ.uuid}
            )

    return new_synd_circs, new_stabilizer_to_circuit


def transversalhadamard(
    interpretation_step: InterpretationStep,
    operation: TransversalHadamard,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Apply a Transversal Hadamard gate to a Block.

    The algorithm is as follows:

    A) CIRCUIT
        - A.1) Apply Hadamard gates over all data qubits of the block.
        - A.2) Append Hadamard gates to the circuit within the InterpretationStep.

    B) STABILIZERS
        - B.1) Update Stabilizers (XX -> ZZ, ZZ -> XX, for 2-body stabilizers) with new UUIDs.
        - B.2) Record the old UUIDs and the new UUIDs of the Stabilizers. (stabilizer_evolution)

    C) PAULIOPERATORS
        - C.1) Update PauliOperators (X to Z Operators) with new UUIDs
               and logical updates.
        - C.2) Update PauliOperators (Z to X Operators) with new UUIDs
               and logical updates.

    D) SYNDROMECIRCUITS
        - D.1) Update Syndrome Circuits with new UUIDs.
        - D.2) Determine the Syndrome Circuit(s) associated with the old Stabilizer(s).
        - D.3) Update the Syndrome Circuit(s) for the new Stabilizer(s).
        - D.4) Update Collection of New Syndrome Circuits and New Stabilizer to Circuit Mapping

    E) NEW BLOCK
        - E.1) Create new Block (Same unique_label as Input Block)

    G) BLOCK HISTORY
        - G.1) Update the latest Block Configuration in InterpretationStep.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the block to apply the transversal hadamard to.
    operation : TransversalHadamard
        Transversal hadamard operation description.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the
        previous operation.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Activating debug mode will enable commutation validation for Block.

    Returns
    -------
    InterpretationStep
        Interpretation step after the transversal hadamard operation.
    """

    # A) CIRCUIT
    # A.1) Apply Hadamard gates over all data qubits of the block.
    block = interpretation_step.get_block(operation.input_block_name)

    data_channels = tuple(
        interpretation_step.get_channel_MUT(str(each_dqbit), "quantum")
        for each_dqbit in block.data_qubits
    )
    hadamard_circ_seq = [
        [Circuit("H", channels=(each_channel,)) for each_channel in data_channels]
    ]

    hadamard_block = Circuit(
        name=f"Transversal Hadamard on {block.unique_label}",
        circuit=hadamard_circ_seq,
    )

    # A.2) Append Hadamard gates to the circuit within the InterpretationStep.
    interpretation_step.append_circuit_MUT(hadamard_block, same_timeslice)

    # B) STABILIZERS
    # Mapping for the effect of the Hadamard gates on the Paulis.
    # Only able to convert X -> Z and Z -> X.
    had_trans = {"X": "Z", "Z": "X"}

    # B.1) Update Stabilizers (XX -> ZZ, ZZ -> XX, for 2-body stabilizers) with new UUIDs.
    new_stabs = tuple(
        Stabilizer(
            pauli="".join(
                had_trans[each_pauli] for each_pauli in each_stabilizer.pauli
            ),
            data_qubits=each_stabilizer.data_qubits,
            ancilla_qubits=each_stabilizer.ancilla_qubits,
        )
        for each_stabilizer in block.stabilizers
        if "Y" not in each_stabilizer.pauli
    )

    if len(new_stabs) != len(block.stabilizers):
        raise ValueError(
            "One of the Stabilizer in the Block contains a Pauli Y, which is not supported by Hadamard."
        )

    # B.2) Record the new UUIDs and the old UUIDs of the Stabilizers. (stabilizer_evolution)
    interpretation_step.stabilizer_evolution.update(
        {
            each_new_stab.uuid: (each_stabilizer.uuid,)
            for each_new_stab, each_stabilizer in zip(
                new_stabs, block.stabilizers, strict=True
            )
        }
    )

    # C) PAULIOPERATORS
    # C.1) Update PauliOperators (X to Z Pauli String in Operators) with new UUIDs
    # and logical updates.

    new_x_pauli_ops = tuple(
        PauliOperator(
            pauli="".join(had_trans[each_pauli] for each_pauli in each_pauli_op.pauli),
            data_qubits=each_pauli_op.data_qubits,
        )
        for each_pauli_op in block.logical_x_operators
    )

    for new_op, old_op in zip(new_x_pauli_ops, block.logical_x_operators, strict=True):
        old_op_updates = interpretation_step.logical_x_operator_updates.get(
            old_op.uuid, ()
        )
        interpretation_step.logical_x_evolution[new_op.uuid] = (old_op.uuid,)
        interpretation_step.update_logical_operator_updates_MUT(
            new_op.pauli[0],
            new_op.uuid,
            new_updates=old_op_updates,
            inherit_updates=False,
        )

    # C.2) Update PauliOperators (Z to X String in Operators) with new UUIDs
    # and logical updates.

    new_z_pauli_ops = tuple(
        PauliOperator(
            pauli="".join(had_trans[each_pauli] for each_pauli in each_pauli_op.pauli),
            data_qubits=each_pauli_op.data_qubits,
        )
        for each_pauli_op in block.logical_z_operators
    )

    for new_op, old_op in zip(new_z_pauli_ops, block.logical_z_operators, strict=True):
        old_op_updates = interpretation_step.logical_z_operator_updates.get(
            old_op.uuid, ()
        )
        interpretation_step.logical_z_evolution[new_op.uuid] = (old_op.uuid,)
        interpretation_step.update_logical_operator_updates_MUT(
            new_op.pauli[0],
            new_op.uuid,
            new_updates=old_op_updates,
            inherit_updates=False,
        )
    # D) SYNDROMECIRCUITS
    #   D.1) Update Syndrome Circuits with new UUIDs.
    #   D.2) Determine the Syndrome Circuit(s) associated with the old Stabilizer(s).
    #   D.3) Update the Syndrome Circuit(s) for the new Stabilizer(s).
    #   D.4) Update Collection of New Syndrome Circuits and New Stabilizer to Circuit Mapping
    new_stabilizer_to_circuit = {}
    new_synd_circ_tuple, new_stabilizer_to_circuit = (
        transversal_hadamard_syndrome_circuits(
            block, interpretation_step, new_stabs, new_stabilizer_to_circuit
        )
    )

    # E) NEW BLOCK
    # E.1) Create new Block (Same unique_label as Input Block)
    new_block = RotatedSurfaceCode(
        stabilizers=new_stabs,
        logical_x_operators=new_z_pauli_ops,
        logical_z_operators=new_x_pauli_ops,
        syndrome_circuits=new_synd_circ_tuple,
        stabilizer_to_circuit=new_stabilizer_to_circuit,
        unique_label=block.unique_label,
        skip_validation=not debug_mode,
    )

    # G) BLOCK HISTORY
    # G.1) Update the latest Block Configuration in InterpretationStep.
    interpretation_step.update_block_history_and_evolution_MUT((new_block,), (block,))

    return interpretation_step



"""
Copyright (c) Entropica Labs Pte Ltd 2025.

Use, distribution and reproduction of this program in its source or compiled
form is prohibited without the express written consent of Entropica Labs Pte
Ltd.

"""

from __future__ import annotations

from functools import cached_property
import textwrap
from uuid import uuid4
from typing import ClassVar

from pydantic import Field, field_validator
from pydantic.dataclasses import dataclass

from .circuit import Circuit
from .channel import Channel, ChannelType
from .utilities import BoolOp
from .utilities.validation_tools import dataclass_config


# pylint: disable=arguments-differ, arguments-renamed, unnecessary-lambda-assignment
@dataclass(config=dataclass_config)
class IfElseCircuit(Circuit):
    """
    Branching circuit: executes if_circuit or else_circuit depending on some classical
    condition circuit.
    """

    # Marker to allow other modules to detect IfElseCircuit instances without
    # importing the class (prevents circular imports). This is a ClassVar so
    # pydantic dataclasses won't treat it as a field.
    _loom_ifelse_marker: ClassVar[bool] = True

    # Define class-specific fields for IfElseCircuit
    if_circuit: Circuit | None = Field(
        default_factory=lambda: Circuit("empty_branch"), validate_default=True
    )
    else_circuit: Circuit | None = Field(
        default_factory=lambda: Circuit("empty_branch"), validate_default=True
    )
    condition_circuit: Circuit | None = Field(
        default_factory=lambda: Circuit(
            name=BoolOp.MATCH,
            channels=[Channel(type=ChannelType.CLASSICAL)],
        ),
        validate_default=True,
    )
    id: str = Field(default_factory=lambda: str(uuid4()))

    # Override fields from Circuit
    name: str = Field(default="if-else_circuit", init=False, frozen=True)
    circuit: tuple[tuple[Circuit, ...], ...] = Field(default_factory=tuple, init=False)
    channels: tuple[Channel, ...] = Field(default_factory=tuple, init=False)

    # Validation functions
    @field_validator("condition_circuit")
    @classmethod
    def validate_condition_circuit(cls, value: Circuit) -> Circuit:
        """
        Provide a default condition circuit if none is provided. Otherwise, check if the
        condition circuit is classical. Throw an error if that is not the case.
        """
        if value is None:
            return Circuit(
                BoolOp.MATCH,
                channels=[Channel(type=ChannelType.CLASSICAL)],
            )

        if not all(channel.is_classical() for channel in value.channels):
            raise ValueError(
                "IfElseCircuit `condition_circuit` must be a circuit with classical "
                f"channels only. Found channels: {value.channels}"
            )

        if value.name in BoolOp.multi_bit_list():
            if len(value.channels) < 2:
                raise ValueError(
                    f"Condition circuit with BoolOp '{value.name}' must have at least "
                    "two classical channels."
                )
        elif value.name in BoolOp.mono_bit_list():
            if len(value.channels) != 1:
                raise ValueError(
                    f"Condition circuit with BoolOp '{value.name}' must have only one "
                    "classical channel."
                )
        else:
            raise ValueError(
                f"Unsupported BoolOp '{value.name}' for condition circuit. Supported "
                "BoolOps are: "
                f"{', '.join(BoolOp.multi_bit_list() + BoolOp.mono_bit_list())}."
            )

        return value

    @field_validator("if_circuit", "else_circuit")
    @classmethod
    def validate_circuit_branches(cls, circuit: Circuit) -> Circuit:
        """
        Assign default empty Circuit if None, and wrap base gates into Circuit if
        needed.
        """
        if circuit is None or circuit.name == "empty_branch":
            return Circuit(name="empty_branch")
        if not circuit.circuit:
            return Circuit(name=f"wrapped_{circuit.name}", circuit=circuit)
        return circuit

    def __post_init__(self):
        """
        Post-initialization to set derived fields. As Circuit objects are immutable,
        we use object.__setattr__ to set these fields. Additionally, we do not perform
        validation again here as the fields are derived from already validated fields.
        """
        # Format circuit field as ((if_circuit, else_circuit),)
        object.__setattr__(self, "circuit", ((self.if_circuit, self.else_circuit),))

        # Gather all unique channels from both branches and prepend condition channels
        # Note that the order of channels is:
        #   condition channels, quantum channels, classical channels
        branch_channels = set(self.if_circuit.channels) | set(
            self.else_circuit.channels
        )
        typing_order = (
            ChannelType.QUANTUM,
            ChannelType.CLASSICAL,
        )
        ordered_branch_channels = tuple(
            sorted(branch_channels, key=lambda x: typing_order.index(x.type))
        )
        all_channels = self.condition_circuit.channels + ordered_branch_channels
        object.__setattr__(self, "channels", all_channels)

        # Set duration to max of branches
        object.__setattr__(
            self, "duration", max(self.if_circuit.duration, self.else_circuit.duration)
        )

    # Override Parent Methods
    @classmethod
    def as_gate(cls):
        """Represent IfElseCircuit as a gate."""
        raise NotImplementedError("IfElseCircuit cannot be represented as a gate.")

    def circuit_seq(self):
        """
        Returns the sequence of sub-circuits in the circuit field.
        """
        raise NotImplementedError(
            "IfElseCircuit cannot be converted into a Circuit sequence."
        )

    def flatten(self) -> IfElseCircuit:
        """
        Flatten the IfElseCircuit by flattening its branches and condition circuit.
        """
        flat_if = self.if_circuit.flatten()
        flat_else = self.else_circuit.flatten()
        flat_condition = self.condition_circuit.flatten()

        return IfElseCircuit(
            if_circuit=flat_if,
            else_circuit=flat_else,
            condition_circuit=flat_condition,
        )

    @classmethod
    def unroll(cls, input_circuit: IfElseCircuit) -> tuple[IfElseCircuit, ...]:

        unrolled_if = input_circuit.if_circuit.unroll(input_circuit.if_circuit)
        unrolled_else = input_circuit.else_circuit.unroll(input_circuit.else_circuit)

        wrapped_unrolled_if = Circuit(
            name=input_circuit.if_circuit.name, circuit=unrolled_if
        )
        wrapped_unrolled_else = Circuit(
            name=input_circuit.else_circuit.name, circuit=unrolled_else
        )

        return (
            IfElseCircuit(
                if_circuit=wrapped_unrolled_if,
                else_circuit=wrapped_unrolled_else,
                condition_circuit=input_circuit.condition_circuit,
            ),
        )

    @cached_property
    def is_condition_single_bit(self) -> bool:
        """Check if the condition circuit is a single-bit condition."""
        return self.condition_circuit.name in BoolOp.mono_bit_list()

    @cached_property
    def is_single_gate_conditioned(self) -> bool:
        """Whether this is just a single gate conditioned by a classical condition."""
        return (
            len(self.if_circuit.circuit) == 1
            and len(self.if_circuit.circuit[0]) == 1
            and not self.if_circuit.circuit[0][0].circuit
            and self.else_circuit.name == "empty_branch"
        )

    def __eq__(self, other: IfElseCircuit) -> bool:
        """Check equality between two IfElseCircuit instances."""
        if not isinstance(other, IfElseCircuit):
            return False
        return (
            self.if_circuit == other.if_circuit
            and self.else_circuit == other.else_circuit
            and self.condition_circuit == other.condition_circuit
        )

    def __repr__(self) -> str:
        """Return a concise string representation of the circuit."""
        return (
            f"{self.name}\n"
            f"  if: {self.if_circuit.name}\n"
            f"  else: {self.else_circuit.name}\n"
            f"  condition: {self.condition_circuit.name}"
        )

    @staticmethod
    def construct_padded_circuit_time_sequence():
        """Construct a padded circuit time sequence."""
        raise NotImplementedError(
            "IfElseCircuit cannot construct a padded circuit time sequence."
        )

    def detailed_str(self) -> str:
        """Return a detailed string representation of the circuit."""
        _skip_firstline = lambda s: s.detailed_str().splitlines()[1:]

        if_str = textwrap.indent("\n".join(_skip_firstline(self.if_circuit)), "    ")
        else_str = textwrap.indent(
            "\n".join(_skip_firstline(self.else_circuit)), "    "
        )

        return (
            f"{self.name}\n"
            f"  if: {self.if_circuit.name}\n"
            f"{if_str}\n"
            f"  else: {self.else_circuit.name}\n"
            f"{else_str}\n"
            f"  condition: {self.condition_circuit.name}\n"
        )



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations

from abc import ABC
from typing import Sequence
from copy import deepcopy
import re

import numpy as np

from .pauli_format_conversion import (
    paulichar_to_xz_npfunc,
    paulixz_to_char_npfunc,
)
from .pauli_commutation import anti_commutes_npfunc
from .pauli_computation import g_npfunc


class PauliOp(ABC):
    """
    Abstract PauliOp class, parent of SignedPauliOp and UnsignedPauliOp.

    Parameters
    ---------
    DTYPE : np.dtype
        The type of numpy arrays used as bit arrays.
    array : np.ndarray
        The array representation of the Pauli operator.
    """

    DTYPE = np.int8
    array: np.ndarray

    @property
    def nqubits(self) -> int:
        """
        The number of qubits that the Pauli operator acts on.
        """
        return len(self.array) // 2

    @property
    def x(self) -> np.array:
        """
        The array representing the X-component of the Pauli operator in binary
        representation.
        """
        return self.array[0 : self.nqubits]

    @property
    def z(self) -> np.array:
        """
        The array representing the Z-component of the Pauli operator in binary
        representation.
        """
        return self.array[self.nqubits : 2 * self.nqubits]

    @property
    def is_trivial(self) -> bool:
        """
        Checks if the Pauli operator is identity.
        """
        return not np.any(self.array)

    def __eq__(self, o_obj: PauliOp) -> bool:
        if not isinstance(o_obj, self.__class__):
            raise TypeError(
                f"Cannot compare {self.__class__} with object of type {type(o_obj)}"
            )
        return np.array_equal(self.array, o_obj.array)


class SignedPauliOp(PauliOp):
    """
    A class describing a SignedPauliOp, a PauliOp operator with a sign that is + or -.

    Parameters
    ----------
    array : np.ndarray | Sequence
        The array representation of the SignedPauliOp.
    """

    # Constructors
    def __init__(self, array: np.ndarray | Sequence, validated: bool = False) -> None:
        """
        Initialization of the SignedPauliOp via a numpy array.

        Parameters
        ----------
        array : np.ndarray | Sequence
            The array representation of the SignedPauliOp
        """
        if not validated:
            if not isinstance(array, (np.ndarray, Sequence)):
                raise TypeError("Input argument should be a NumPy array or a sequence.")

            if not isinstance(array, np.ndarray):
                # make it into an array
                array = np.array(array)

            if array.ndim != 1:
                raise ValueError("NumPy array should be 1-D.")

            if not len(array) % 2 == 1:
                raise ValueError("Numpy array has to have an odd number of bits")

            # check if all elements are 0 or 1
            if not np.all(array * (array - 1) == 0):
                raise ValueError("The input array has to consist of 0 and 1.")

            if not array.dtype == self.DTYPE:
                # cast it into the correct data type
                array = array.astype(self.DTYPE)

        self.array = array

    @classmethod
    def from_string(cls, pauli_str: str) -> SignedPauliOp:
        """
        Create a SignedPauliOp from a Pauli string, like "+IXZZY"

        Parameters
        ----------
        pauli_str : str
            The Pauli string to create the SignedPauliOp from.

        Returns
        -------
        SignedPauliOp
            The SignedPauliOp created from the Pauli string.

        Raises
        ------
        ValueError
            If the first character of the Pauli string is not '+' or '-'.
        """

        sign = pauli_str[0]
        pauli_chars = pauli_str[1:]

        nqubits = len(pauli_chars)
        array = np.zeros(2 * nqubits + 1, dtype=cls.DTYPE)

        if sign == "+":
            pass
        elif sign == "-":
            array[-1] = 1
        else:
            raise ValueError(
                "The first character of the a SignedPauliOp string should be '+' or  '-"
            )

        # set the array values from
        x, z = paulichar_to_xz_npfunc(np.array(list(pauli_chars)))
        array[0:nqubits] = x
        array[nqubits : 2 * nqubits] = z

        return cls(array, validated=True)

    @classmethod
    def from_sparse_string(
        cls, pauli_str: str, nqubits: int | None = None
    ) -> SignedPauliOp:
        """
        Create a SignedPauliOp from a sparse Pauli string, like "+X2Z5Y7"

        Parameters
        ----------
        pauli_str : str
            The sparse Pauli string to create the SignedPauliOp from. It is not case
            sensitive.
        nqubits : int | None
            The total number of qubits the Pauli operator acts on. Since SignedPauliOp
            is a dense representation, all qubits with index starting from 0 until
            nqubits-1 will be described by the array. The number of qubits should be
            greater than the maximum index in the Pauli string. If None, the number of
            qubits is inferred from the maximum index in the Pauli string.

        Returns
        -------
        SignedPauliOp
            The SignedPauliOp created from the Pauli string.

        Raises
        ------
        ValueError
            If the first character of the Pauli string is not '+' or '-'.
            If there are invalid elements in the Pauli string.
            If the indices in the Pauli string are not unique.
            If the number of qubits is not greater than the maximum index in the Pauli
            string.
        """
        # Get the the Pauli index pairs
        pauli_index_pairs = pauli_str[1:]

        # Parse the sign
        match pauli_str[0]:
            case "+":
                sign = 0
            case "-":
                sign = 1
            case _:
                raise ValueError(
                    "The first character of the a Pauli string should be '+' or '-'."
                )

        # Define the regular expression pattern
        pattern = re.compile(r"([XYZ])(\d+)", re.IGNORECASE)

        # Find all matches in the input string
        matches = pattern.findall(pauli_index_pairs)

        # Check for invalid segments that do not match the pattern
        invalid_segments = pattern.sub("", pauli_index_pairs)
        if invalid_segments:
            raise ValueError(
                f"Invalid elements in the Pauli string: {invalid_segments}."
            )

        # Get the indices and the pauli operators into 2 separate arrays
        indices = np.array([int(idx) for _, idx in matches])
        paulis = np.array([op for op, _ in matches])

        # check that the indices are unique
        if len(indices) != len(set(indices)):
            repeating_indices = {
                int(idx) for idx in indices if list(indices).count(idx) > 1
            }
            raise ValueError(
                f"Qubit indices {repeating_indices} appear more than once in the "
                "Pauli string."
            )

        # Get the maximum index
        max_index = max(indices)

        # Infer the number of qubits if not provided
        if nqubits is None:
            nqubits = max_index + 1
        else:
            if nqubits < max_index + 1:
                raise ValueError(
                    f"Qubit index {max_index} is out of range for {nqubits} qubits."
                )
        # Convert the Pauli operators to x and z bit values
        x_vals, z_vals = paulichar_to_xz_npfunc(paulis)

        # Initialize the array
        array = np.zeros(2 * nqubits + 1, dtype=cls.DTYPE)
        # Set the array values from the sparse string
        array[indices] = x_vals
        array[nqubits + indices] = z_vals
        array[-1] = sign

        return cls(array, validated=True)

    @classmethod
    def identity(cls, nqubits: int, negative=False) -> SignedPauliOp:
        """
        Create an identity operator acting on nqubits. Can be used to create a
        negative identity operator as well.

        Parameters
        ----------
        nqubits : int
            The number of qubits the identity operator should act on.
        negative : bool
            If True, return a negative identity operator.

        Returns
        -------
        SignedPauliOp
            The identity operator acting on nqubits.
        """
        array = np.zeros(2 * nqubits + 1, dtype=cls.DTYPE)
        if negative:
            array[-1] = 1
        return SignedPauliOp(array, validated=True)

    # Magic methods
    def __str__(self) -> str:
        """
        The string representation of the SignedPauliOp
        """
        sign = "+" if self.array[-1] == 0 else "-"
        return sign + "".join(paulixz_to_char_npfunc(self.x, self.z))

    def __repr__(self) -> str:
        """
        The representation of the SignedPauliOp.
        """
        return f"{self.__class__.__name__}({str(self)})"

    def __mul__(self, other: SignedPauliOp) -> SignedPauliOp:
        """
        Multiply two SignedPauliOps.

        Parameters
        ----------
        other : SignedPauliOp
            The other SignedPauliOp to multiply with.

        Returns
        -------
        SignedPauliOp
            The product of the two SignedPauliOps.
        """
        if not isinstance(other, SignedPauliOp):
            raise TypeError("Can only multiply with another SignedPauliOp.")

        if self.nqubits != other.nqubits:
            raise ValueError(
                "The two Pauli operators should act on the same number of qubits."
            )

        if pauliops_anti_commute(self, other):
            raise ValueError(
                "Cannot multiply anti-commuting Pauli operators. That "
                "would give an imaginary unit contribution in the product."
                "To multiply anti-commuting operators use the method "
                "multiply_with_anticommuting_operator instead."
            )

        # Get the product of the 2 Pauli operators without the sign
        x = self.x ^ other.x
        z = self.z ^ other.z

        # If the sum of g modulo 4 is:
        # 0 -> positive sign (0) contribution
        # 2 -> negative sign (1) contribution
        # 3 and 1 are only possible for anti-commuting Pauli operators.
        sign = (
            self.sign
            ^ other.sign
            ^ sum(g_npfunc(self.x, self.z, other.x, other.z)) % 4 // 2
        )

        return SignedPauliOp(
            np.concatenate((x, z, [sign])).astype(self.DTYPE), validated=True
        )

    # Properties
    @property
    def sign(self) -> SignedPauliOp.DTYPE:
        """
        The sign of the Pauli operator, either 0 or 1.
        """
        return self.array[-1]

    @property
    def is_minus_identity(self) -> bool:
        """
        Check if the Pauli operator is a minus identity operator. Returns True if the
        Pauli operator is a minus identity operator, False otherwise.
        """
        return np.all(self.array[:-1] == 0) and self.array[-1] == 1

    # Methods
    def as_sparse_string(self) -> str:
        """
        Return the Pauli operator as a sparse Pauli string.

        Returns
        -------
        str
            The Pauli operator as a sparse Pauli string.
        """
        # Find the indices of the non-zero elements
        x_indices = np.nonzero(self.x)[0]
        z_indices = np.nonzero(self.z)[0]
        indices = np.union1d(x_indices, z_indices)
        # Get the Pauli operators
        paulis = paulixz_to_char_npfunc(self.x[indices], self.z[indices])
        # Get the sign
        sign = "+" if self.sign == 0 else "-"
        # Combine the sign and the Pauli operators
        return sign + "".join(
            f"{pauli}{idx}" for idx, pauli in zip(indices, paulis, strict=True)
        )

    def with_flipped_sign(self) -> SignedPauliOp:
        """
        Return a copy of the operator with the sign flipped.

        Returns
        -------
        SignedPauliOp
            The Pauli operator with the sign flipped.
        """
        return SignedPauliOp(
            np.concatenate((self.array[:-1], [1 - self.sign])).astype(self.DTYPE),
            validated=True,
        )

    def reindexed(
        self, qubit_map: list[int], nqubits: int | None = None
    ) -> SignedPauliOp:
        """
        Return a copy of the operator with the qubits reindexed according to the
        qubit map.

        Parameters
        ----------
        qubit_map : list[int]
            The mapping of the new qubit indices to the old qubit indices.
        nqubits : int | None
            The total number of qubits the Pauli operator acts on. If None, the number
            of qubits is inferred from the maximum index in the qubit map.

        Returns
        -------
        SignedPauliOp
            The Pauli operator with the reindexed qubits.
        """
        if nqubits is None:
            nqubits = max(qubit_map) + 1

        if len(set(qubit_map)) != len(qubit_map):
            raise ValueError("The qubit map should not contain duplicate indices.")

        if len(qubit_map) != self.nqubits:
            raise ValueError(
                f"The qubit map length ({len(qubit_map)}) should be equal "
                f"to the number of qubits in the Pauli operator ({self.nqubits})."
            )

        if nqubits <= max(qubit_map):
            raise ValueError(
                f"The number of qubits ({nqubits}) should be greater than "
                f"the maximum index in the qubit map ({max(qubit_map)})."
            )
        # cast the qubit map to a numpy array
        qubit_map = np.array(qubit_map)
        # make the new array
        new_array = np.zeros(2 * nqubits + 1, dtype=self.DTYPE)
        new_array[-1] = self.sign
        new_array[qubit_map] = self.x
        new_array[nqubits + qubit_map] = self.z

        return SignedPauliOp(new_array, validated=True)

    # Methods
    def multiply_with_anticommuting_operator(
        self, other: SignedPauliOp
    ) -> SignedPauliOp:
        """
        Multiply two Pauli operators that anti-commute by including a factor of i.
        For 2 Pauli operators A and B that anti-commute,
        A.multiply_with_anticommuting_operator(B) returns the SignedPauliOp i * A * B.

        Parameters
        ----------
        other : SignedPauliOp
            The other Pauli operator to multiply with.

        Returns
        -------
        SignedPauliOp
            The product of the two Pauli operators with a factor of i.
        """
        if not isinstance(other, SignedPauliOp):
            raise TypeError("Can only multiply with another SignedPauliOp.")

        if self.nqubits != other.nqubits:
            raise ValueError(
                "The two Pauli operators should act on the same number of qubits."
            )

        if not pauliops_anti_commute(self, other):
            raise ValueError(
                "Can only multiply anti-commuting Pauli operators. "
                "To multiply commuting operator use the operator *"
            )

        # Find the exponent of the imaginary unit
        i_exp = (
            sum(g_npfunc(self.x, self.z, other.x, other.z))  # from the commutation #
            + 2 * self.sign  # from the sign of the first operator #
            + 2 * other.sign  # from the sign of the second operator #
            + 1  # from the imaginary unit #
        ) % 4  # i^4 = 1 #
        # And then deduce the sign
        sign = i_exp // 2

        # Get the product of the 2 Pauli operators without the sign
        x = self.x ^ other.x
        z = self.z ^ other.z

        return SignedPauliOp(
            np.concatenate((x, z, [sign])).astype(self.DTYPE), validated=True
        )

    def copy(self) -> SignedPauliOp:
        """
        Returns a deep copy of the SignedPauliOp object.

        Returns
        -------
        SignedPauliOp
            Deep copy of the SignedPauliOp object
        """
        return deepcopy(self)


def pauliops_anti_commute(op1: PauliOp, op2: PauliOp) -> int:
    """
    Given two pauli strings, find their anti-commutation value.
    A value of `0` means that they commute and a value of `1` means that they
    anti-commute.

    Parameters
    ----------
    op1 : PauliOp
        One of the Pauli operators.
    op2 : PauliOp
        The other Pauli operator.

    Returns
    -------
    int
        The anti-commutation value of the two Pauli operators.
        0 if they commute, 1 if they anti-commute.
    """
    if not (isinstance(op1, PauliOp) and isinstance(op2, PauliOp)):
        raise TypeError("Both inputs should be of type PauliOp.")

    anti_comm_array = anti_commutes_npfunc(op1.x, op1.z, op2.x, op2.z)

    return np.sum(anti_comm_array) % 2


class UnsignedPauliOp(PauliOp):
    """A class describing an UnsignedPauliOp, a Pauli operator without a sign."""

    def __init__(self, array: np.ndarray | Sequence, validated: bool = False) -> None:
        """Initialization of the UnsignedPauliOp via a numpy array.

        Parameters
        ----------
        array : np.ndarray | Sequence
            The array representation of the UnsignedPauliOp
        """
        if not validated:
            if not isinstance(array, (np.ndarray, Sequence)):
                raise TypeError("Input argument should be a NumPy array or a sequence.")

            if not isinstance(array, np.ndarray):
                # make it into an array
                array = np.array(array)

            if array.ndim != 1:
                raise ValueError("NumPy array should be 1-D.")

            if not len(array) % 2 == 0:
                raise ValueError("Numpy array has to have an even number of bits")

            # check if all elements are 0 or 1
            if not np.all(array * (array - 1) == 0):
                raise ValueError("The input array has to consist of 0 and 1.")

            if not array.dtype == self.DTYPE:
                # cast it into the correct data type
                array = array.astype(self.DTYPE)

        self.array = array

    @classmethod
    def from_string(cls, pauli_str: str) -> UnsignedPauliOp:
        """Create an UnsignedPauliOp from a Pauli string, like "IXZZY"

        Parameters
        ----------
        pauli_str : str
            The Pauli string to create the UnsignedPauliOp from.

        Returns
        -------
        UnsignedPauliOp
            The UnsignedPauliOp created from the Pauli string.
        """
        if pauli_str[0] in ["+", "-"]:
            raise ValueError(
                "The first character of the a UnsignedPauliOp cannot be '+' or '-'. "
                "Maybe you want to use SignedPauliOp class instead."
            )
        pauli_chars = pauli_str
        nqubits = len(pauli_chars)
        array = np.zeros(2 * nqubits, dtype=cls.DTYPE)

        # set the array values from
        x, z = paulichar_to_xz_npfunc(np.array(list(pauli_chars)))
        array[0:nqubits] = x
        array[nqubits : 2 * nqubits] = z

        return cls(array, validated=True)

    def __str__(self) -> str:
        """The string representation of the UnSignedPauliOp"""
        return "".join(paulixz_to_char_npfunc(self.x, self.z))

    def copy(self) -> UnsignedPauliOp:
        """
        Returns a deep copy of the UnsignedPauliOp object.
        """
        return deepcopy(self)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from ..eka import Circuit, Block, Stabilizer

from .circuit_validation import is_circuit_valid
from .debug_dataclass import DebugData
from .utilities import logical_states_to_check, logical_state_transformations_to_check


def is_syndrome_extraction_circuit_valid(
    circuit: Circuit,
    input_block: Block | tuple[Block, ...],
    measurement_to_input_stabilizer_map: dict[str, Stabilizer],
) -> DebugData:
    """Tests if the syndrome extraction circuit is a valid one.
    The function checks that the circuit:
    - does not alter the check operators
    - acts on the logical level as identity
    - measures indeed the specified input check operators with specific measurement
    operations

    Parameters
    ----------
    circuit : Circuit
        The syndrome extraction circuit.
    input_block : Block | tuple[Block, ...]
        The input Block object(s).
    measurement_to_input_stabilizer_map : dict[str, Stabilizer]
        Dictionary matching the classical channel name of a measurement operation with a
        stabilizer in the input code.

    Returns
    -------
    DebugData
        The result of the checks.
    """

    logical_state_transformations = [
        (state, (state,))
        for state in logical_states_to_check(input_block.n_logical_qubits)
    ]
    # Check if the circuit is valid
    return is_circuit_valid(
        circuit=circuit,
        input_block=input_block,
        output_block=input_block,
        output_stabilizers_parity={},
        output_stabilizers_with_any_value=[],
        logical_state_transformations_with_parity={},
        logical_state_transformations=logical_state_transformations,
        measurement_to_input_stabilizer_map=measurement_to_input_stabilizer_map,
    )


def is_logical_operation_circuit_valid(
    circuit: Circuit,
    input_block: Block | tuple[Block, ...],
    x_operators_sparse_pauli_map: list[str],
    z_operators_sparse_pauli_map: list[str],
) -> DebugData:
    """Checks if the logical operation circuit is a valid one.
    The function checks that the circuit:
    - does not alter the check operators
    - acts on the logical level in a way that is consistent with the logical operation
    defined by the x and z maps

    For example, for a CNOT gate from qubit 0 to qubit 1 we know that:
    - X0 -> X0X1
    - X1 -> X1
    - Z0 -> Z0
    - Z1 -> Z0Z1

    and thus the input pauli map arguments will be:
    - `x_operators_sparse_pauli_map = ["X0X1", "X1"]`
    - `z_operators_sparse_pauli_map = ["Z0", "Z0Z1"]`

    Parameters
    ----------
    circuit : Circuit
        The logical operation circuit.
    input_block : Block | tuple[Block, ...]
        The input Block object(s).
    x_operators_sparse_pauli_map : list[str]
        The list of sparse Pauli strings describing how each X operator is transformed.
        No sign is needed and the order matches the transformation of the logical
        operators X0, X1, ...
    z_operators_sparse_pauli_map : list[str]
        The list of sparse Pauli strings describing how each Z operator is transformed.
        No sign is needed and the order matches the transformation of the logical
        operators Z0, Z1, ...

    Returns
    -------
    DebugData
        The result of the checks.
    """
    # Get the logical state transformations to check
    logical_state_transformations = logical_state_transformations_to_check(
        x_operators_sparse_pauli_map, z_operators_sparse_pauli_map
    )
    return is_circuit_valid(
        circuit=circuit,
        input_block=input_block,
        output_block=input_block,
        output_stabilizers_parity={},
        output_stabilizers_with_any_value=[],
        logical_state_transformations_with_parity={},
        logical_state_transformations=logical_state_transformations,
        measurement_to_input_stabilizer_map={},
    )



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from functools import partial

from pydantic.dataclasses import dataclass
from pydantic import Field, field_validator


from .base_operation import Operation
from .logical_measurement import LogicalMeasurement
from ..utilities import (
    SingleQubitPauliEigenstate,
    Direction,
    Orientation,
    ResourceState,
    dataclass_config,
    larger_than_zero_error,
)


# CodeOperation act on the code itself
@dataclass(config=dataclass_config)
class CodeOperation(Operation):
    """
    Parent class for all code operations. All code operations act on blocks

    Properties
    ----------

    _inputs : tuple[str, ...]
        Standardized way to access the input blocks names.
    _outputs : tuple[str, ...]
        Standardized way to access the output blocks names.
    """

    @property
    def _inputs(self):
        """
        Standardized way to access the input block(s) names.

        Returns
        -------
        tuple[str, ...]
            Names of the input blocks
        """
        if hasattr(self, "input_block_name"):
            return (self.input_block_name,)
        if hasattr(self, "input_blocks_name"):
            return self.input_blocks_name

        raise ValueError(f"No block inputs specified for {self.__class__.__name__}")

    @property
    def _outputs(self):
        """
        Standardized way to access the output block(s) names.

        Returns
        -------
        tuple[str, ...]
            Names of the output blocks
        """
        if hasattr(self, "output_block_name"):
            return (self.output_block_name,)
        if hasattr(self, "output_blocks_name"):
            return self.output_blocks_name

        return self._inputs


# Readout operations
@dataclass(config=dataclass_config)
class MeasureBlockSyndromes(CodeOperation):
    """
    Performs a given number of rounds of syndrome measurements on a block.

    Parameters
    ----------
    input_block_name : str
        Name of the block to measure.
    n_cycles : int
        Number of cycles to measure. Default is 1.
    """

    input_block_name: str
    n_cycles: int = 1


@dataclass(config=dataclass_config)
class MeasureLogicalX(CodeOperation):
    """
    Measure the logical X operator of a block.

    Parameters
    ----------
    input_block_name : str
        Name of the block where the logical operator to be measured is located.
    logical_qubit : int | None, optional
        Index of the logical qubit inside the specified block which should be measured.
        For blocks with a single logical qubit, this parameter does not need to be
        provided. Then by default the index 0 is chosen for this single logical qubit.
        For blocks with multiple logical qubits, this parameter is required.
    """

    input_block_name: str
    logical_qubit: int = Field(default=0)


@dataclass(config=dataclass_config)
class MeasureLogicalZ(CodeOperation):
    """
    Measure the logical Z operator of a block.

    Parameters
    ----------
    input_block_name : str
        Name of the block where the logical operator to be measured is located.
    logical_qubit : int | None, optional
        Index of the logical qubit inside the specified block which should be measured.
        For blocks with a single logical qubit, this parameter does not need to be
        provided. Then by default the index 0 is chosen for this single logical qubit.
        For blocks with multiple logical qubits, this parameter is required.
    """

    input_block_name: str
    logical_qubit: int = Field(default=0)


@dataclass(config=dataclass_config)
class MeasureLogicalY(CodeOperation):
    """
    Measure the logical Y operator of a block.

    Parameters
    ----------
    input_block_name : str
        Name of the block where the logical operator to be measured is located.
    logical_qubit : int | None, optional
        Index of the logical qubit inside the specified block which should be measured.
        For blocks with a single logical qubit, this parameter does not need to be
        provided. Then by default the index 0 is chosen for this single logical qubit.
        For blocks with multiple logical qubits, this parameter is required.
    """

    input_block_name: str
    logical_qubit: int = Field(default=0)


@dataclass(config=dataclass_config)
class LogicalX(CodeOperation):
    """
    Apply a logical X operator to a block.

    Parameters
    ----------
    input_block_name : str
        Name of the block where the logical operator should be applied.
    logical_qubit : int | None, optional
        Index of the logical qubit inside the specified block which should be acted on.
        For blocks with a single logical qubit, this parameter does not need to be
        provided. Then by default the index 0 is chosen for this single logical qubit.
        For blocks with multiple logical qubits, this parameter is required.
    """

    input_block_name: str
    logical_qubit: int = Field(default=0)


@dataclass(config=dataclass_config)
class LogicalY(CodeOperation):
    """
    Apply a logical Y operator to a block.

    Parameters
    ----------
    input_block_name : str
        Name of the block where the logical operator should be applied.
    logical_qubit : int | None, optional
        Index of the logical qubit inside the specified block which should be acted on.
        For blocks with a single logical qubit, this parameter does not need to be
        provided. Then by default the index 0 is chosen for this single logical qubit.
        For blocks with multiple logical qubits, this parameter is required.
    """

    input_block_name: str
    logical_qubit: int = Field(default=0)


@dataclass(config=dataclass_config)
class LogicalZ(CodeOperation):
    """
    Apply a logical Z operator to a block.

    Parameters
    ----------
    input_block_name : str
        Name of the block where the logical operator should be applied.
    logical_qubit : int | None, optional
        Index of the logical qubit inside the specified block which should be acted on.
        For blocks with a single logical qubit, this parameter does not need to be
        provided. Then by default the index 0 is chosen for this single logical qubit.
        For blocks with multiple logical qubits, this parameter is required.
    """

    input_block_name: str
    logical_qubit: int = Field(default=0)


@dataclass(config=dataclass_config)
class ResetAllDataQubits(CodeOperation):
    """
    Reset all data qubits to a specific SingleQubitPauliEigenstate.

    input_block_name : str
        Name of the block where the logical operator should be applied.
    state: SingleQubitPauliEigenstate | None, optional
        State to which the logical qubit should be reset. Default is
        SingleQubitPauliEigenstate.ZERO, i.e. the zero eigenstate of the Pauli Z
        operator.
    """

    input_block_name: str
    state: SingleQubitPauliEigenstate = Field(default=SingleQubitPauliEigenstate.ZERO)


@dataclass(config=dataclass_config)
class ResetAllAncillaQubits(CodeOperation):
    """
    Reset all ancilla qubits to a specific SingleQubitPauliEigenstate.

    Parameters
    ----------
    input_block_name : str
        Name of the block where the logical operator should be applied.
    state: SingleQubitPauliEigenstate | None, optional
        State to which the ancilla qubit should be reset. Default is
        SingleQubitPauliEigenstate.ZERO, i.e. the zero eigenstate of the Pauli Z
        operator.
    """

    input_block_name: str
    state: SingleQubitPauliEigenstate = Field(default=SingleQubitPauliEigenstate.ZERO)


@dataclass(config=dataclass_config)
class Grow(CodeOperation):
    """
    Grow operation.

    Parameters
    ----------
    input_block_name : str
        Name of the block to grow.
    direction : Direction
        Direction in which to grow the block.
    length : int
        Length by which to grow the block.
    """

    input_block_name: str
    direction: Direction
    length: int

    _position_error = field_validator("length", mode="before")(
        partial(larger_than_zero_error, arg_name="length")
    )


@dataclass(config=dataclass_config)
class Shrink(CodeOperation):
    """
    Shrink operation.

    Parameters
    ----------
    input_block_name : str
        Name of the block to shrink.
    direction : Direction
        Direction in which to shrink the block.
    length : int
        Length by which to shrink the block.
    """

    input_block_name: str
    direction: Direction
    length: int
    _position_error = field_validator("length", mode="before")(
        partial(larger_than_zero_error, arg_name="length")
    )


@dataclass(config=dataclass_config)
class Merge(CodeOperation):
    """
    Merge operation.

    Parameters
    ----------
    input_blocks_name : tuple[str, str]
        Names of the two blocks to merge.
    output_block_name : str
        Name of the resulting block.
    orientation : Orientation, optional
        Orientation along which to merge the blocks. E.g. if Orientation.HORIZONTAL,
        the blocks will be merged using their left and right boundaries (whichever is
        easiest). If None, the orientation will be derived from the blocks positions.
    """

    input_blocks_name: tuple[str, str]
    output_block_name: str
    orientation: Orientation | None = Field(default=None, validate_default=True)


@dataclass(config=dataclass_config)
class Split(CodeOperation):
    """
    Split operation.

    Parameters
    ----------
    input_block_name : str
        Name of the block to split.
    output_blocks_name : tuple[str, str]
        Names of the resulting blocks.
    orientation : Orientation
        Orientation along which to split the block. E.g. if Orientation.HORIZONTAL, the
        block will be split in a horizontal cut, leaving two blocks with adjacent top
        and bottom boundaries.
    split_position : int
        Position at which to split the block, distance to the (0,0) corner of the block.
    """

    input_block_name: str
    output_blocks_name: tuple[str, str]
    orientation: Orientation
    split_position: int
    _position_error = field_validator("split_position", mode="before")(
        partial(larger_than_zero_error, arg_name="split_position")
    )


@dataclass(config=dataclass_config)
class StateInjection(CodeOperation):
    """
    Inject the given resource state into the specified block. This operation resets the
    central qubit of the block into the specified resource state and maximizes the
    number of stabilizers that can be initialized in a deterministic way.

    E.g. a T state injection in a RotatedSurfaceCode block will reset the central qubit
    into the T state and reset the rest of the data qubits into four quadrants,
    such that two quadrants are in the ``|0⟩`` state and two quadrants are in the
    ``|+⟩`` state. This ensures that the Z stabilizer measurements are deterministic in
    the ``|0⟩`` quadrants and the X stabilizer measurements are deterministic in the
    ``|+⟩`` quadrants.

    Parameters
    ----------
    input_block_name : str
        Name of the block to inject the resource state into.
    resource_state : ResourceState
        The resource state to inject into the block. This can be one of the following:
        - ResourceState.T: T state
        - ResourceState.S: S state
    """

    input_block_name: str
    resource_state: ResourceState


@dataclass(config=dataclass_config)
class ConditionalLogicalX(CodeOperation):
    """
    Apply a conditional logical X operator to a block.

    Parameters
    ----------
    input_block_name : str
        Name of the block where the logical operator should be applied.
    logical_qubit : int | None, optional
        Index of the logical qubit inside the specified block which should be acted on.
        For blocks with a single logical qubit, this parameter does not need to be
        provided. Then by default the index 0 is chosen for this single logical qubit.
        For blocks with multiple logical qubits, this parameter is required.
    condition : LogicalMeasurement | None, optional
        Condition for logical pauli operation to be applied based on the value of the
        LogicalMeasurement provided.
    """

    input_block_name: str
    condition: LogicalMeasurement
    logical_qubit: int = Field(default=0)


@dataclass(config=dataclass_config)
class ConditionalLogicalY(CodeOperation):
    """
    Apply a conditional logical Y operator to a block.

    Parameters
    ----------
    input_block_name : str
        Name of the block where the logical operator should be applied.
    logical_qubit : int | None, optional
        Index of the logical qubit inside the specified block which should be acted on.
        For blocks with a single logical qubit, this parameter does not need to be
        provided. Then by default the index 0 is chosen for this single logical qubit.
        For blocks with multiple logical qubits, this parameter is required.
    condition : LogicalMeasurement
        Condition for logical pauli operation to be applied based on the value of the
        LogicalMeasurement provided.
    """

    input_block_name: str
    condition: LogicalMeasurement
    logical_qubit: int = Field(default=0)


@dataclass(config=dataclass_config)
class ConditionalLogicalZ(CodeOperation):
    """
    Apply a conditional logical Z operator to a block.

    Parameters
    ----------
    input_block_name : str
        Name of the block where the logical operator should be applied.
    logical_qubit : int | None, optional
        Index of the logical qubit inside the specified block which should be acted on.
        For blocks with a single logical qubit, this parameter does not need to be
        provided. Then by default the index 0 is chosen for this single logical qubit.
        For blocks with multiple logical qubits, this parameter is required.
    condition : LogicalMeasurement
        Condition for logical pauli operation to be applied based on the value of the
        LogicalMeasurement provided.
    """

    input_block_name: str
    condition: LogicalMeasurement
    logical_qubit: int = Field(default=0)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import numpy as np
import networkx as nx


def binary_gaussian_elimination(matrix: np.ndarray) -> np.ndarray:
    """
    Performs binary gaussian elimination on an input matrix.

    Algorithm reference:
    https://www.cs.umd.edu/~gasarch/TOPICS/factoring/fastgauss.pdf

    Implementation:
    https://gist.github.com/popcornell/bc29d1b7ba37d824335ab7b6280f7fec
    """
    matrix = matrix.copy()

    nrows, ncols = matrix.shape

    # Deal with trivial cases
    if nrows == 0 or ncols == 0:
        return matrix

    i = 0
    # Each iteration of this loop will ensure that column j does not contain any 1s
    # from row i downwards.
    # If possible, row i will contain the only 1 in column j. This is possible if there
    # is a row with index pivot_row >= i that has a 1 in column j.
    for j in range(ncols):
        # Find index of first `1` in column j, from row i-th onwards.
        pivot_row = np.argmax(matrix[i:, j]) + i

        # If the pivot is 0, skip this column
        # The reason is that in this column j no row below the current row has a 1
        # in it (thus the pivot row has a zero in column j), so we can't use this
        # column to eliminate 1s in other rows.
        if matrix[pivot_row, j] == 0:
            continue

        # Swap pivot with row i
        matrix[[pivot_row, i]] = matrix[[i, pivot_row]]

        # XOR the pivot row that we just relocated into the i-th row into all other rows
        # to make all other entries in the column j 0
        ## Target row i, from column j onwards
        target_row = matrix[i, j:]

        target_column = np.copy(matrix[:, j])
        target_column[i] = 0  # don't XOR the pivot row with itself

        # The array with which we will XOR the matrix
        # Every row in column j that is 1 will be XORed with the target row
        xor_mask = np.outer(target_column, target_row)

        # XOR the matrix
        matrix[:, j:] = matrix[:, j:] ^ xor_mask

        # Increment row index
        i += 1
        # If we reach the last row, break the loop
        if i >= nrows:
            break

    return matrix


def find_maximum_matching(graph: nx.Graph) -> dict[int, int]:
    """Obtains maximum matching of a bipartite graph.

    Parameters
    ----------
    graph : nx.Graph
        Networkx Graph from which to extract the maximum matching. The graph needs to
        be bipartite.

    Returns
    ------
    matching: dict[int,int]
        Matching as dictionary whose values and keys contain the matching connections
        between nodes.
    """

    # Ensure the graph is bipartite
    if not nx.algorithms.bipartite.is_bipartite(graph):
        raise ValueError("Graph is not bipartite.")

    # Dictionary to hold the matching results
    matching = {}

    # Iterate over connected components
    for component in nx.connected_components(graph):
        # Extract the subgraph for the current component
        subgraph = graph.subgraph(component)

        # Find the bipartite sets for the component
        _, top_nodes = nx.algorithms.bipartite.sets(subgraph)

        # Find the maximum matching for the component using Hopcroft-Karp algorithm
        local_matching = nx.algorithms.bipartite.maximum_matching(
            subgraph, top_nodes=top_nodes
        )

        # Add to the overall matching dictionary
        matching.update(local_matching)

    return matching


def minimum_edge_coloring(graph: nx.Graph) -> dict[int, list[tuple[int, int]]]:
    """
    Computes the minimum edge coloring of a bipartite graph ``graph``. The chromatic
    index for all bipartite graphs is equal to the maximum degree of the graph. In the
    context of leveraging this algorithm for circuit construction, this is equivalent to
    distributing more gates into a single layer.

    Parameters
    ----------
    graph : nx.Graph
        A bipartite graph.

    Returns
    -------
    coloring: dict[int,list]
        Minimum edge coloring as a dictionary where keys are color indices and values
        the list of edges.
    """

    # Get the maximum degree of the graph, corresponding to the number of colors needed
    d = max(dict(graph.degree()).values())

    # Initialize coloring dictionary
    coloring = {i: [] for i in range(d)}
    incident_color_tracker = {node: set() for node in graph.nodes()}

    sorted_edges = [tuple(sorted(edge)) for edge in graph.edges]

    # Assign colors, ranging from 0 to d-1
    for edge in sorted_edges:
        u, v = edge
        used_colors = incident_color_tracker[u] | incident_color_tracker[v]
        color = next(i for i in range(d) if i not in used_colors)

        coloring[color].append(edge)
        incident_color_tracker[u].add(color)
        incident_color_tracker[v].add(color)

    return coloring


def extract_subgraphs_from_edge_labels(
    graph: nx.Graph, label_attribute: str = "cardinality"
) -> dict[str, nx.Graph]:
    """
    Extract subgraphs of the Tanner Graphs induced by the edges associated with a
    particular attribute. This function is mainly used for obtaining the ones associated
    with the cardinality decorators in the cardinal circuit constructor function.

    Parameters
    ----------
    graph : nx.Graph
        Tanner Graph from HGP code whose edges are decorated according to the different
        attributes.

    label_attribute: str
        Edge attribute name. Defaults to cardinality.

    Returns
    -------
    subgraphs_dict: dict[str,nx.Graph]
        Dictionary containing the value of each attribute as a key and the associated
        induced subgraph as a value.
    """

    # Check input
    if not isinstance(label_attribute, str):
        raise TypeError("Label attribute must be a string.")

    attribute_names = {key for _, _, attr in graph.edges(data=True) for key in attr}
    if label_attribute not in attribute_names:
        raise ValueError(f"Edge attribute {label_attribute} not present in graph.")

    # Check for any unlabelled edges w.r.t. to label_attribute input
    unlabelled_edges = [
        (a, b)
        for a, b, info in graph.edges(data=True)
        if info.get(label_attribute) is None
    ]
    if len(unlabelled_edges) > 0:
        raise ValueError(f"Edges {unlabelled_edges} do not contain input label.")

    # Extract unique edge labels
    attribute_values = set(nx.get_edge_attributes(graph, label_attribute).values())

    # Initialize subgraph dictionary
    subgraphs_dict = {}

    # Create subgraphs for each label
    for value in attribute_values:
        # Filter edges with the current label
        edges_with_value = [
            (u, v) for u, v, d in graph.edges(data=True) if d[label_attribute] == value
        ]

        # Create a subgraph with those edges
        subgraph = graph.edge_subgraph(edges_with_value).copy()

        # Store the subgraph in the dictionary
        subgraphs_dict[value] = subgraph

    return subgraphs_dict


def cardinality_distribution(t_graph: nx.Graph) -> dict[str, list]:
    """
    Extracts the edges in a HGP Tanner Graph associated with each cardinality {N,E,S,W}.

    Parameters
    ----------
    t_graph: nx.Graph
        Tanner graph associated with HGP code.

    Returns
    -------
    edge_dict: dict[str,list]
        A dictionary whose keys are the different cardinalities and the values
        all the edges associated with it.

    """

    # Check input
    attribute_names = {key for _, _, attr in t_graph.edges(data=True) for key in attr}

    if len(attribute_names - {"cardinality"}) > 0:
        raise ValueError(
            f"Only allowed attribute name is 'cardinality', but input "
            f"contains: {attribute_names-{'cardinality'}}."
        )

    edge_labels = set(nx.get_edge_attributes(t_graph, "cardinality").values())

    if len(edge_labels - {"E", "N", "S", "W"}) > 0:
        raise ValueError(
            f"Only cardinal values 'E','N','S','W' are allowed attributes, "
            f"but input contains invalid ones: {edge_labels-{'E', 'N', 'S', 'W'}}."
        )

    # Check for any unlabelled edges
    unlabelled_edges = {(a, b) for a, b, info in t_graph.edges(data=True) if info == {}}

    if len(unlabelled_edges) > 0:
        raise ValueError(f"Edges {unlabelled_edges} are not labeled.")

    # Initialize dictionary
    edge_dict = {}

    # Extract all the edges associated with each cardinality
    for label in edge_labels:
        edge_dict[label] = [
            (u, v) for u, v, d in t_graph.edges(data=True) if d["cardinality"] == label
        ]

    return edge_dict


def verify_css_code_condition(hx: np.ndarray, hz: np.ndarray) -> bool:
    """
    Verifies that the parity-check matrices define a valid CSS code, by checking that
    the X and Z check matrices are orthogonal, i.e. that the stabilizers commute.

    Parameters
    ----------
    hx : np.ndarray
        X parity-check matrix.
    hz : np.ndarray
        Z parity-check matrix.

    Returns
    -------
    valid: bool
        If True, the parity-check matrices define a valid CSS code.
    """

    # Check input
    for h in [hx, hz]:
        if not np.all(np.isin(h, [0, 1])):
            raise ValueError("Parity-check matrix contains non-binary elements.")

        if not np.any(h):
            raise ValueError("Parity-check matrix is empty.")

    # Compute the symplectic product of the parity-check matrices
    product = np.dot(hx, hz.T) % 2

    # Specify validity as a boolean
    valid = not bool(product.any())

    return valid



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations
from enum import Enum

from loom.eka.utilities.enums import enum_missing


class FourBodySchedule(str, Enum):
    """
    Enum for the four-body schedule used in the rotated surface code.
    """

    N = "N"
    Z = "Z"

    def opposite_schedule(self) -> FourBodySchedule:
        """Get the opposite schedule."""
        if self == FourBodySchedule.N:
            return FourBodySchedule.Z
        if self == FourBodySchedule.Z:
            return FourBodySchedule.N
        raise ValueError("Invalid schedule. Cannot determine opposite schedule.")

    @classmethod
    def _missing_(cls, value):
        """
        Allow inputs with upper-case characters. For more details, see the
        documentation of ``enum_missing`` at the beginning of the file.
        """
        return enum_missing(cls, value)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# pylint: disable=duplicate-code, too-many-lines
from __future__ import annotations
from uuid import uuid4
from functools import cached_property
from itertools import combinations
from pydantic.dataclasses import dataclass

from loom.eka import (
    Block,
    Lattice,
    LatticeType,
    PauliOperator,
    Stabilizer,
    SyndromeCircuit,
    Circuit,
    Channel,
)
from loom.eka.utilities import (
    Direction,
    Orientation,
    DiagonalDirection,
    dataclass_config,
)

from loom_rotated_surface_code.utilities import FourBodySchedule


# pylint: disable=duplicate-code
@dataclass(config=dataclass_config)
class RotatedSurfaceCode(Block):  # pylint: disable=too-many-public-methods
    """
    A sub-class of ``Block`` that represents a rotated surface code block.
    Contains methods to create a rotated surface code block along with
    properties to access the block's size, upper left qubit, stabilizers,
    logical operators, and other relevant information.
    """

    @classmethod
    def create(
        # pylint: disable=too-many-branches,too-many-statements, too-many-locals
        # pylint: disable=too-many-arguments, too-many-positional-arguments
        cls,
        dx: int,
        dz: int,
        lattice: Lattice,
        unique_label: str | None = None,
        position: tuple[int, ...] = (0, 0),
        x_boundary: Orientation = Orientation.HORIZONTAL,
        weight_2_stab_is_first_row: bool = True,
        weight_4_x_schedule: FourBodySchedule | None = None,
        logical_x_operator: PauliOperator | None = None,
        logical_z_operator: PauliOperator | None = None,
        skip_validation: bool = False,
    ) -> RotatedSurfaceCode:
        """
        Create a ``Block`` object for a rotated surface code block. The orientation of
        the block (i.e. where which boundaries are) and where the weight-2 stabilizers
        are can be controlled with the ``x_boundary`` and ``weight_2_stab_is_first_row``
        argument. By default, the top row and the left column are chosen as the logical
        operators. Their pauli string and whether they belong to the logical Z or X
        operator depends on the orientation of the boundaries.

        The coordinates used for data qubits are the following (here as an example for a
        d=3 rotated surface code, with ``x_boundary=Orientation.H`` and
        ``weight_2_stab_is_first_row=True``):

        ..code-block::

                               Z
              (0,0) --- (1,0) --- (2,0)
                |         |         |
            X   |    Z    |    X    |
                |         |         |
              (0,1) --- (1,1) --- (2,1)
                |         |         |
                |    X    |    Z    |  X
                |         |         |
              (0,2) --- (1,2) --- (2,2)
                     Z

        Parameters
        ----------
        dx : int
            Size of the block in the horizontal direction
        dz : int
            Size of the block in the vertical direction
        lattice : Lattice
            Lattice on which the block is defined. The qubit indices depend on the type
            of lattice.
        unique_label : str, optional
            Label for the block. It must be unique among all blocks in the initial Eka.
            If no label is provided, a unique label is generated automatically using the
            uuid module.
        position : tuple[int, ...], optional
            Position of the top left corner of the block on the lattice, by default
            (0, 0)
        x_boundary : Orientation, optional
            Specifies whether the X boundaries are horizontal (Orientation.HORIZONTAL),
            i.e. going from left to right, or vertical (Orientation.V), i.e. going from
            top to bottom.
            The X boundary is the boundary that exhibits X Pauli charges. In other words
            it is the boundary with 2-body Z stabilizers. By default
            Orientation.HORIZONTAL.
        weight_2_stab_is_first_row : bool, optional
            Specifies whether the top most weight-2 stabilizer at the left boundary is
            in the first row (if True) or in the second row (if False), by default True
        weight_4_x_schedule : FourBodySchedule, optional
            Schedule for measuring the XXXX stabilizer, by default None. If None is
            provided, the schedule is calculated from the orientation of the X
            boundary. E.g. if ``x_boundary`` is Orientation.HORIZONTAL, the schedule is
            set to FourBodySchedule.N.
            The default scheme is described in https://arxiv.org/abs/1404.3747 III, B.
            In the example above, ``weight_4_x_schedule`` is ``FourBodySchedule.N``,
            this is equivalent to measure the upper right XXXX stabilizer in the order
            (2,0) -> (2,1) -> (1,0) -> (1,1).
        logical_x_operator: PauliOperator | None, optional
            Logical X operator. If None is provided, by default the top row or the left
            column is chosen (depending on the orientation of the block as specified by
            the ``x_boundary`` and ``weight_2_stab_is_first_row`` parameter)
        logical_z_operator: PauliOperator | None, optional
            Logical Z operator. If None is provided, by default the top row or the left
            column is chosen (depending on the orientation of the block as specified by
            the ``x_boundary`` and ``weight_2_stab_is_first_row`` parameter)
        skip_validation : bool, optional
            Skip validation of the block object, by default False.

        Returns
        -------
        Block
            Block object for a rotated surface code block
        """

        # Input validation
        if lattice.lattice_type != LatticeType.SQUARE_2D:
            raise ValueError(
                "The creation of rotated surface code blocks is "
                "currently only supported for 2D square lattices. Instead "
                f"the lattice is of type {lattice.lattice_type}."
            )

        if not isinstance(position, tuple) or any(
            not isinstance(x, int) for x in position
        ):
            raise ValueError(
                f"`position` must be a tuple of integers. Got '{position}' instead."
            )

        # Check that lattice co-ordinate of the qubit == system dimension
        # The unit vector is not included in the lattice co-ordinate.
        if len(position) != lattice.n_dimensions:
            raise ValueError(
                f"`position` has length {len(position)} while length "
                f"{lattice.n_dimensions} is required to match the lattice dimension."
            )

        if unique_label is None:
            unique_label = str(uuid4())

        if isinstance(x_boundary, Orientation) is False:
            x_boundary = Orientation(x_boundary)

        # Create stabilizers
        top_left_is_xxxx = (
            x_boundary == Orientation.HORIZONTAL
        ) != weight_2_stab_is_first_row
        # Create the schedule for stabilizers
        if weight_4_x_schedule is None:
            weight_4_x_schedule = (
                FourBodySchedule.N
                if x_boundary == Orientation.HORIZONTAL
                else FourBodySchedule.Z
            )
        elif isinstance(weight_4_x_schedule, str):
            weight_4_x_schedule = FourBodySchedule(weight_4_x_schedule)
        weight_4_z_schedule = weight_4_x_schedule.opposite_schedule()

        # Generate weight-4 stabilizers covering half of the block in a checkerboard
        # pattern starting in the top left corner
        if top_left_is_xxxx:
            weight4_stabs_top_left = cls.generate_weight4_stabs(
                dx, dz, "XXXX", weight_4_x_schedule, True
            )
        else:
            weight4_stabs_top_left = cls.generate_weight4_stabs(
                dx, dz, "ZZZZ", weight_4_z_schedule, True
            )
        # Generate the remaining weight-4 stabilizers
        if top_left_is_xxxx:
            weight4_stabs_others = cls.generate_weight4_stabs(
                dx, dz, "ZZZZ", weight_4_z_schedule, False
            )
        else:
            weight4_stabs_others = cls.generate_weight4_stabs(
                dx, dz, "XXXX", weight_4_x_schedule, False
            )

        stab_left_right_is_x = top_left_is_xxxx != weight_2_stab_is_first_row
        stab_left_right = "XX" if stab_left_right_is_x else "ZZ"
        stab_top_bottom = "ZZ" if stab_left_right_is_x else "XX"

        # Left boundary
        if dz % 2 == 1:
            num_weight2_stabs = (dz - 1) / 2
        else:
            num_weight2_stabs = dz / 2 - (not weight_2_stab_is_first_row)
        stabs_left = cls.generate_weight2_stabs(
            pauli=stab_left_right,
            initial_position=(0, (not weight_2_stab_is_first_row)),
            num_stabs=num_weight2_stabs,
            orientation=Orientation.VERTICAL,
            is_bottom_or_right=False,
        )

        # Right boundary
        right_first_row = weight_2_stab_is_first_row != (dx % 2)
        if dz % 2 == 1:
            num_weight2_stabs = (dz - 1) / 2
        else:
            num_weight2_stabs = dz / 2 - (not right_first_row)
        stabs_right = cls.generate_weight2_stabs(
            pauli=stab_left_right,
            initial_position=(dx - 1, (not right_first_row)),
            num_stabs=num_weight2_stabs,
            orientation=Orientation.VERTICAL,
            is_bottom_or_right=True,
        )

        # Top boundary
        num_weight2_stabs = dx // 2
        if dx % 2 == 1:
            num_weight2_stabs = (dx - 1) / 2
        else:
            num_weight2_stabs = dx / 2 - weight_2_stab_is_first_row
        stabs_top = cls.generate_weight2_stabs(
            pauli=stab_top_bottom,
            initial_position=(weight_2_stab_is_first_row, 0),
            num_stabs=num_weight2_stabs,
            orientation=Orientation.HORIZONTAL,
            is_bottom_or_right=False,
        )

        # Bottom boundary
        bottom_first_col = weight_2_stab_is_first_row == (dz % 2)
        if dx % 2 == 1:
            num_weight2_stabs = (dx - 1) / 2
        else:
            num_weight2_stabs = dx / 2 - (not bottom_first_col)
        stabs_bottom = cls.generate_weight2_stabs(
            pauli=stab_top_bottom,
            initial_position=((not bottom_first_col), dz - 1),
            num_stabs=num_weight2_stabs,
            orientation=Orientation.HORIZONTAL,
            is_bottom_or_right=True,
        )

        # Combine all stabilizers
        stabilizers = (
            weight4_stabs_top_left
            + weight4_stabs_others
            + stabs_left
            + stabs_right
            + stabs_top
            + stabs_bottom
        )

        # Define syndrome circuits
        xxxx_syndrome_circuit = cls.generate_syndrome_circuit(
            pauli="XXXX", padding=None, name="XXXX"
        )
        zzzz_syndrome_circuit = cls.generate_syndrome_circuit(
            pauli="ZZZZ", padding=None, name="ZZZZ"
        )
        # Get left padding and generate left syndrome circuit
        left_padding = cls.find_padding(
            boundary="left",
            schedule=(
                weight_4_x_schedule if stab_left_right_is_x else weight_4_z_schedule
            ),
        )
        left_syndrome_circuit = cls.generate_syndrome_circuit(
            pauli=stab_left_right,
            padding=left_padding,
            name=f"left-{stab_left_right}",
        )
        # Get right padding and generate right syndrome circuit
        right_padding = cls.find_padding(
            boundary="right",
            schedule=(
                weight_4_x_schedule if stab_left_right_is_x else weight_4_z_schedule
            ),
        )
        right_syndrome_circuit = cls.generate_syndrome_circuit(
            pauli=stab_left_right,
            padding=right_padding,
            name=f"right-{stab_left_right}",
        )
        # Get top padding and generate top syndrome circuit
        top_padding = cls.find_padding(
            boundary="top",
            schedule=(
                weight_4_z_schedule if stab_left_right_is_x else weight_4_x_schedule
            ),
        )
        top_syndrome_circuit = cls.generate_syndrome_circuit(
            pauli=stab_top_bottom,
            padding=top_padding,
            name=f"top-{stab_top_bottom}",
        )
        # Get bottom padding and generate bottom syndrome circuit
        bottom_padding = cls.find_padding(
            boundary="bottom",
            schedule=(
                weight_4_z_schedule if stab_left_right_is_x else weight_4_x_schedule
            ),
        )
        bottom_syndrome_circuit = cls.generate_syndrome_circuit(
            pauli=stab_top_bottom,
            padding=bottom_padding,
            name=f"bottom-{stab_top_bottom}",
        )
        # Construct the list of all syndrome circuits
        syndrome_circuits = [
            xxxx_syndrome_circuit,
            zzzz_syndrome_circuit,
            top_syndrome_circuit,
            bottom_syndrome_circuit,
            left_syndrome_circuit,
            right_syndrome_circuit,
        ]

        # Create stabilizer_to_circuit mapping
        stabilizer_to_circuit = (
            {
                stab.uuid: (
                    xxxx_syndrome_circuit.uuid
                    if top_left_is_xxxx
                    else zzzz_syndrome_circuit.uuid
                )
                for stab in weight4_stabs_top_left
            }
            | {
                stab.uuid: (
                    zzzz_syndrome_circuit.uuid
                    if top_left_is_xxxx
                    else xxxx_syndrome_circuit.uuid
                )
                for stab in weight4_stabs_others
            }
            | {stab.uuid: left_syndrome_circuit.uuid for stab in stabs_left}
            | {stab.uuid: right_syndrome_circuit.uuid for stab in stabs_right}
            | {stab.uuid: top_syndrome_circuit.uuid for stab in stabs_top}
            | {stab.uuid: bottom_syndrome_circuit.uuid for stab in stabs_bottom}
        )

        # Create logical operators
        if logical_x_operator is None:
            qubits = (
                [(dx_i, 0, 0) for dx_i in range(dx)]
                if x_boundary == Orientation.HORIZONTAL
                else [(0, dz_i, 0) for dz_i in range(dz)]
            )
            logical_x_operator = PauliOperator(
                pauli="X" * len(qubits), data_qubits=qubits
            )

        if logical_z_operator is None:
            qubits = (
                [(0, dz_i, 0) for dz_i in range(dz)]
                if x_boundary == Orientation.HORIZONTAL
                else [(dx_i, 0, 0) for dx_i in range(dx)]
            )
            logical_z_operator = PauliOperator(
                pauli="Z" * len(qubits), data_qubits=qubits
            )

        block = cls(
            unique_label=unique_label,
            stabilizers=stabilizers,
            logical_x_operators=[logical_x_operator],
            logical_z_operators=[logical_z_operator],
            syndrome_circuits=syndrome_circuits,
            stabilizer_to_circuit=stabilizer_to_circuit,
            skip_validation=skip_validation,
        )
        if position == (0, 0):
            return block

        return block.shift(position)

    @staticmethod
    def find_padding(
        boundary: Direction, schedule: FourBodySchedule
    ) -> tuple[int, int]:
        """
        Finds the padding indices for the two body stabilizers. Padding indices
        are used to indicate empty time steps in the syndrome circuits. This allows
        to standardize the size of syndrome circuits for the rotated surface code.

        Parameters
        ----------
        boundary : Direction
            Type of boundary for the surface code can be LEFT, RIGHT, TOP, or
                BOTTOM
        schedule : FourBodySchedule
            Schedule for measuring the four body stabilizers, we can deduce the two
                body schedule from this.

        Returns
        -------
        tuple[int, int]
            Padding indices.
        """
        match (schedule, boundary):
            case (FourBodySchedule.N, Direction.LEFT) | (
                FourBodySchedule.Z,
                Direction.BOTTOM,
            ):
                return (2, 3)
            case (FourBodySchedule.N, Direction.RIGHT) | (
                FourBodySchedule.Z,
                Direction.TOP,
            ):
                return (0, 1)
            case (FourBodySchedule.N, Direction.TOP) | (
                FourBodySchedule.Z,
                Direction.RIGHT,
            ):
                return (0, 2)
            case (FourBodySchedule.N, Direction.BOTTOM) | (
                FourBodySchedule.Z,
                Direction.LEFT,
            ):
                return (1, 3)
            case _:
                return ValueError(
                    f"The boundary {boundary} and schedule {schedule} are"
                    f" not compatible. Only N and Z schedules are "
                    "supported."
                )

    @staticmethod
    def generate_syndrome_circuit(
        pauli: str, padding: tuple[int, ...] | None, name: str
    ) -> SyndromeCircuit:
        """
        Generates a syndrome circuit for a given Pauli string. The syndrome
        circuits generated are all the same size, where padding indices indicate
        where to add empty time steps.

        Parameters
        ----------
        pauli : str
            Pauli string associated the stabilizer
        padding : tuple[int, int] | None
            Padding indices for the two body stabilizers. Padding indices are used
            to locate empty spaces in the syndrome circuits.
        name : str
            Name of the syndrome circuit

        Returns
        -------
        SyndromeCircuit
            Syndrome circuit for the Pauli string
        """
        weight = len(pauli)
        data_channels = [Channel(type="quantum", label=f"d{i}") for i in range(weight)]
        cbit_channel = Channel(type="classical", label="c0")
        ancilla_channel = Channel(type="quantum", label="a0")

        reset = [Circuit("Reset_0", channels=[ancilla_channel])]
        hadamard1 = [Circuit("H", channels=[ancilla_channel])]
        hadamard2 = [Circuit("H", channels=[ancilla_channel])]
        # If there is no padding, there are no extra empty time steps
        if padding is None:
            entangle_ancilla = [
                [Circuit(f"C{p}", channels=[ancilla_channel, data_channels[i]])]
                for i, p in enumerate(pauli)
            ]
        # If there is padding, we add empty time steps to the circuit
        else:
            # We create lists of single items to insert None at the padding indices
            padded_pauli = list(pauli)
            padded_data_channels = list(data_channels)
            for i in padding:
                padded_pauli.insert(i, None)
                padded_data_channels.insert(i, None)

            entangle_ancilla = [
                (
                    [
                        Circuit(
                            f"C{p}",
                            channels=[ancilla_channel, padded_data_channels[i]],
                        )
                    ]
                    if p is not None  # If not None, we add the empty time step
                    else []
                )
                for i, p in enumerate(padded_pauli)
            ]
        measurement = [Circuit("Measurement", channels=[ancilla_channel, cbit_channel])]

        circuit_list = [reset, hadamard1] + entangle_ancilla + [hadamard2, measurement]
        return SyndromeCircuit(
            pauli=pauli,
            name=name,
            circuit=Circuit(
                name=name,
                circuit=circuit_list,
                channels=data_channels + [ancilla_channel, cbit_channel],
            ),
        )

    # Properties

    @property
    def size(self) -> tuple[int, int]:
        """
        Return the size of the block in the horizontal and vertical direction.
        """
        size_0 = max(qb[0] for qb in self.data_qubits) - min(
            qb[0] for qb in self.data_qubits
        )
        size_1 = max(qb[1] for qb in self.data_qubits) - min(
            qb[1] for qb in self.data_qubits
        )
        return (size_0 + 1, size_1 + 1)

    @property
    def upper_left_qubit(self) -> tuple[int, ...]:
        """
        Return the qubit with the smallest coordinates in the block.
        """
        return min(self.data_qubits, key=lambda x: x[0] + x[1])

    @property
    def upper_left_4body_stabilizer(self) -> Stabilizer:
        """
        Return the 4-body stabilizer associated with the upper left qubit.
        """
        return [
            stab
            for stab in self.stabilizers
            if len(stab.data_qubits) == 4 and self.upper_left_qubit in stab.data_qubits
        ][0]

    # NOTE these four properties should be deleted and included as fields at the
    # creation of the RotatedSurfaceCode instance
    @property
    def weight_4_x_schedule(self) -> FourBodySchedule:
        """
        Return the schedule for measuring the XXXX stabilizer.
        """
        for stab in self.stabilizers:
            if stab.pauli == "XXXX":
                schedule = (
                    FourBodySchedule.N
                    if stab.data_qubits[1][1] > stab.data_qubits[0][1]
                    else FourBodySchedule.Z
                )
                return schedule
        raise ValueError("XXXX stabilizer not found in the stabilizers.")

    @property
    def weight_4_z_schedule(self) -> FourBodySchedule:
        """
        Return the schedule for measuring the ZZZZ stabilizer.
        """
        return (
            FourBodySchedule.N
            if self.weight_4_x_schedule == FourBodySchedule.Z
            else FourBodySchedule.Z
        )

    @property
    def x_boundary(self) -> Orientation:
        """
        Return the orientation of the X boundary.
        """
        if self.boundary_type(Direction.TOP) == "X":
            return Orientation.HORIZONTAL

        # if the top boundary is Z, then the X boundary is vertical
        return Orientation.VERTICAL

    @property
    def weight_2_stab_is_first_row(self) -> bool:
        """
        Return whether the top most weight-2 stabilizer at the left boundary is
        in the first row.
        """
        first_qubit = self.upper_left_qubit
        second_qubit = tuple(
            q if i != 1 else q + 1 for i, q in enumerate(self.upper_left_qubit)
        )
        return any(
            # first_qubit in stab.data_qubits and second_qubit in stab.data_qubits
            # and len(stab.data_qubits) == 2
            set([first_qubit, second_qubit]) == set(stab.data_qubits)
            for stab in self.stabilizers
        )

    @property
    def topological_corners(self) -> tuple[tuple[int, ...], ...]:
        """
        Return the coordinates of the topological corners of the block.
        """
        return tuple(q for q, p in self.pauli_charges.items() if p == "Y")

    @property
    def geometric_corners(self) -> tuple[tuple[int, ...], ...]:
        """
        Return the coordinates of the geometric corners of the block. The geometric
        corners that can be detected are qubits that may be the single most:
        - top-left qubit
        - top-right qubit
        - bottom-left qubit
        - bottom-right qubit
        - right-qubit
        - bottom-qubit
        - left-qubit
        - top-qubit
        """

        def get_sole_max(lambda_expression):
            """
            Find and return the qubit with the maximum value of the lambda expression
            only if it is the only qubit with that value. If there are multiple qubits
            with the same value, return None.
            """
            max_value = max(lambda_expression(q) for q in self.data_qubits)
            max_qubits = [
                q for q in self.data_qubits if lambda_expression(q) == max_value
            ]
            return max_qubits[0] if len(max_qubits) == 1 else None

        lambdas_to_maximize = [
            lambda x: x[0] + x[1],  # bottom right corner
            lambda x: x[0] - x[1],  # top right corner
            lambda x: -x[0] + x[1],  # bottom left corner
            lambda x: -x[0] - x[1],  # top left corner
            lambda x: x[0],  # right corner
            lambda x: x[1],  # bottom corner
            lambda x: -x[0],  # left corner
            lambda x: -x[1],  # top corner
        ]

        return tuple(
            get_sole_max(lambda_expression)
            for lambda_expression in lambdas_to_maximize
            if get_sole_max(lambda_expression) is not None
        )

    @property
    def all_boundary_stabilizers(self) -> tuple[Stabilizer, ...]:
        """
        Return the stabilizers associated with any boundary.
        """
        return (
            self.boundary_stabilizers(direction=Direction.TOP)
            + self.boundary_stabilizers(direction=Direction.BOTTOM)
            + self.boundary_stabilizers(direction=Direction.LEFT)
            + self.boundary_stabilizers(direction=Direction.RIGHT)
        )

    @property
    def bulk_stabilizers(self) -> tuple[Stabilizer, ...]:
        """
        Return the stabilizers not associated with any boundary.
        """
        return tuple(
            stab
            for stab in self.stabilizers
            if stab not in self.all_boundary_stabilizers
        )

    @property
    def orientation(self) -> Orientation | None:
        """
        Return the orientation of the block. If the block is square, return None.
        """
        if self.size[0] > self.size[1]:
            return Orientation.HORIZONTAL
        if self.size[0] < self.size[1]:
            return Orientation.VERTICAL
        return None

    @property
    def is_horizontal(self) -> bool:
        """
        Return True if the horizontal size is larger than the vertical size.
        """
        return self.orientation == Orientation.HORIZONTAL

    @property
    def is_vertical(self) -> bool:
        """
        Return True if the vertical size is larger than the horizontal size.
        """
        return self.orientation == Orientation.VERTICAL

    # Static methods
    @staticmethod
    def generate_weight4_stabs(
        dx: int,
        dz: int,
        pauli: str,
        schedule: FourBodySchedule,
        start_in_top_left_corner: bool,
        initial_position: tuple[int, int] = (0, 0),
    ) -> list[Stabilizer]:
        """
        Generate the list of all weight-4 stabilizers of a rotated surface code
        block of the given type (= pauli string).

        Parameters
        ----------
        dx: int
            Distance in the horizontal direction. If dx=3, there will be 2 weight-4
            stabilizers created in the horizontal direction.
        dz: int
            Distance in the vertical direction. If dz=3, there will be 2 weight-4
            stabilizers created in the vertical direction.
        pauli : str
            Pauli string of the stabilizers
        start_in_top_left_corner : bool
            If True, the first stabilizer will start in the top left corner and then
            follow the alternating checkerboard pattern. If False, it will be
            exactly the opposite covering of the checkerboard pattern.
        schedule : FourBodySchedule
            Schedule for measuring the four body stabilizers, see
            https://arxiv.org/abs/1404.3747, III, B. for more details.
        initial_position : tuple[int, int], optional
            Initial position where the chain of weight-4 stabilizers should start,
            by default (0, 0)

        Returns
        -------
        list[Stabilizer]
            List of weight-4 stabilizers of the specified type for the rotated
            surface code
        """
        if len(initial_position) > 2:
            initial_position = initial_position[:2]
        elif len(initial_position) < 2:
            raise ValueError(
                "Initial position must be a tuple of length >= 2. Got "
                f"{initial_position}."
            )
        x, z = initial_position
        is_n_pattern = schedule == FourBodySchedule.N
        return [
            Stabilizer(
                pauli=pauli,
                data_qubits=[
                    (x + dx_i + 1, z + dz_i, 0),
                    (x + dx_i + is_n_pattern, z + dz_i + is_n_pattern, 0),
                    (x + dx_i + (not is_n_pattern), z + dz_i + (not is_n_pattern), 0),
                    (x + dx_i, z + dz_i + 1, 0),
                ],
                ancilla_qubits=[(x + dx_i + 1, z + dz_i + 1, 1)],
            )
            for dx_i in range(dx - 1)
            for dz_i in range(dz - 1)
            if (dx_i + dz_i) % 2 != start_in_top_left_corner
        ]

    @staticmethod
    def generate_weight2_stabs(
        pauli: str,
        initial_position: tuple[int, int],
        num_stabs: int,
        orientation: Orientation,
        is_bottom_or_right: bool,
    ) -> list[Stabilizer]:
        """
        Generate the list of all weight-2 stabilizers along one of the four
        boundaries. Note that the schedule for measuring the weight-4 stabilizers
        does not change the order in which we specify the weight-2 stabilizers.
        For stabilizers along the right and bottom boundaries, the stabilizers are
        generated differently. The stabilizers along the right and bottom boundary have
        the coordinates of the ancilla qubits shifted by (1, 0) and (0, 1) respectively.

        Parameters
        ----------
        pauli : str
            Pauli string of the stabilizers
        initial_position : tuple[int, int]
            Initial position where the chain of weight-2 stabilizers should start
        num_stabs : int
            Number of weight-2 stabilizers to generate along this boundary
        orientation : Orientation
            Orientation of the boundary, either HORIZONTAL or VERTICAL
        is_bottom_or_right : bool
            If True, the stabilizers are along the bottom or right boundaries, this
            means their ancilla qubit is 'outside' of the block geometry. If False, the
            stabilizers are along the top or left boundaries.

        Returns
        -------
        list[Stabilizer]
            List of weight-2 stabilizers along the specified boundary
        """
        is_horizontal = orientation == Orientation.HORIZONTAL
        return [
            Stabilizer(
                pauli=pauli,
                data_qubits=[
                    (
                        initial_position[0] + is_horizontal * (2 * i + 1),
                        initial_position[1] + (not is_horizontal) * 2 * i,
                        0,
                    ),
                    (
                        initial_position[0] + is_horizontal * (2 * i),
                        initial_position[1] + (not is_horizontal) * (2 * i + 1),
                        0,
                    ),
                ],
                ancilla_qubits=[
                    (
                        initial_position[0]
                        + 1 * (not is_horizontal) * is_bottom_or_right
                        + is_horizontal * (2 * i + 1),
                        initial_position[1]
                        + 1 * is_horizontal * is_bottom_or_right
                        + (not is_horizontal) * (2 * i + 1),
                        1,
                    )
                ],
            )
            for i in range(0, int(num_stabs))
        ]

    # Instance methods

    def boundary_qubits(self, direction: Direction | str) -> list[tuple[int, ...]]:
        """
        Return the data qubits that are part of the specified boundary.

        Parameters
        ----------
        direction : Direction | str
            Boundary (top, bottom, left, or right) for which the data qubits should be
            returned. If a string is provided, it is converted to a Direction enum.

        Returns
        -------
        list[tuple[int, ...]]
            Data qubits that are part of the specified boundary
        """
        # Input validation: cast direction to Direction enum if it is not already
        if not isinstance(direction, Direction):
            direction = Direction(direction)

        axis = 1 if direction in [Direction.TOP, Direction.BOTTOM] else 0
        selector_function = (
            max if direction in [Direction.BOTTOM, Direction.RIGHT] else min
        )
        min_or_max_value = selector_function(qb[axis] for qb in self.data_qubits)
        return [qb for qb in self.data_qubits if qb[axis] == min_or_max_value]

    def boundary_type(self, direction: Direction | str) -> str:
        """
        Return the type of the specified boundary, either X or Z. This assumes that
        the block is a standard rotated surface code block which is a square block with
        Y charges at the four corners and X and Z charges at the boundaries.

        Note that there are different conventions about when to call a boundary X or Z.
        We call a boundary X type if it exhibits X Pauli charges. In other words, it is
        of X type if the stabilizers along the boundary are Z stabilizers.

        Parameters
        ----------
        direction : Direction | str
            Boundary (top, bottom, left, or right) for which the data qubits should be
            returned. If a string is provided, it is converted to a Direction enum.

        Returns
        -------
        str
            Type of the boundary, either X or Z
        """
        # Input validation: cast direction to Direction if it is not already
        if not isinstance(direction, Direction):
            direction = Direction(direction)

        # Input validation: check that block size is > 2 in the direction of the
        # boundary
        if direction in [Direction.TOP, Direction.BOTTOM] and self.size[1] <= 2:
            raise ValueError(
                "The block is too small to have a top or bottom boundary of well "
                "defined type."
            )
        if direction in [Direction.LEFT, Direction.RIGHT] and self.size[0] <= 2:
            raise ValueError(
                "The block is too small to have a left or right boundary of well "
                "defined type."
            )

        # Get the Pauli charges of the boundary qubits,
        # transform them into a set to remove duplicates,
        # and remove the "Y" charge to be left with the X and Z charges.
        # For the standard rotated surface code block, there should be only one charge
        # left.
        boundary_pauli_charges = list(
            set(self.pauli_charges[qb] for qb in self.boundary_qubits(direction))
            - set("Y")
        )
        if len(boundary_pauli_charges) > 1:
            raise RuntimeError(
                "Boundary has multiple Pauli charges. This should not happen for the "
                "standard rotated surface code block. The boundary pauli charges "
                f"(excluding Y charges) are {boundary_pauli_charges}."
            )

        return boundary_pauli_charges[0]

    def boundary_stabilizers(self, direction: Direction) -> tuple[Stabilizer, ...]:
        """
        Return the stabilizers associated with the given boundary direction.
        """
        boundary_qubits = self.boundary_qubits(direction)
        return tuple(
            stab
            for stab in self.stabilizers
            if all(q in boundary_qubits for q in stab.data_qubits)
        )

    def get_corner_from_direction(
        self, which_corner: DiagonalDirection
    ) -> tuple[int, int, int]:
        """
        Get the coordinates of the qubit at the specified corner of the block.
        """
        return (
            set(self.boundary_qubits(which_corner.components[0]))
            .intersection(self.boundary_qubits(which_corner.components[1]))
            .pop()
        )

    @cached_property
    def config_and_pivot_corners(
        self,
    ) -> tuple[int, tuple[tuple[int, int, int], ...]]:
        """
        Classify the Block based on the geometry of its topological corners. Return the
        config and the topological corners ordered in a specific way that reflects their
        geometric arrangement.

        Type 1: Rectangular config
                Four topological corners coincide with 4 geometric corners
                The list of corners is returned in the order:
                (top-left, bottom-left, bottom-right, top-right)

                Visualized example::

                    1-------4
                    |       |
                    |       |
                    |       |
                    2-------3

        Type 2: U-config
                Block has rectangular shape with size (d, 2d-1) or (2d-1, d)
                Three of four topological corners coincide with three geometric corners,
                the last topological corner resides on the middle of the long edge whose
                ends occupy only one topological corner
                The list of corners is returned in the order:
                (long_end, middle_edge, angle, short_end)

                Visualized example (can be rotated)::

                    1-------|
                    |       |
                    |       2
                    |       |
                    3-------4

        Type 3: L-config
                Block has rectangular shape with size (d, 2d-1) or (2d-1, d)
                Three of four topological corners coincide with three geometric corners,
                the last topological corner resides on the middle of the long edge whose
                ends occupy two topological corners.
                The list of corners is returned in the order as seen below:
                (long_end, middle_edge, angle, short_end)

                Visualized example (can be rotated)::

                    1-------|
                    |       |
                    2       |
                    |       |
                    3-------4

        Type 4: U-config (phase gate)
                Block has rectangular shape with size (d, 2d) or (2d, d)
                Three of four topological corners coincide with three geometric corners,
                the last topological corner resides in the middle of the long edge whose
                ends occupy only one topological corner. Since the length of a long edge
                is even, there are two qubits near the middle point of the edge.
                Only support the case where the qubit is further away from the end
                occupied by a topological corner.
                The list of corners is returned in the order:
                (long_end, middle_edge, angle, short_end)

                Visualized example (can be rotated)::

                    1-------|
                    |       |
                    |       2
                    |       |
                    |       |
                    3-------4

        Type 0: Other configs

        Returns
        -------
        int:
            The configuration type of the block.
        tuple[tuple[int, int, int], ...]:
            The list of corners in the order specified by the configuration type.
        """
        # Find topological corners that are also geometric corners
        dx, dz = self.size
        is_block_horizontal = dx >= dz
        topological_geometric_corners = set(self.topological_corners) & set(
            self.geometric_corners
        )

        if len(topological_geometric_corners) == 4:
            # Type 1 (rectangle)
            return 1, sorted(topological_geometric_corners)

        is_almost_two_squares = dx == 2 * dz - 1 or dz == 2 * dx - 1
        is_two_squares = dx == 2 * dz or dz == 2 * dx
        is_ul_config = len(topological_geometric_corners) == 3 and is_almost_two_squares
        is_uphase_config = len(topological_geometric_corners) == 3 and is_two_squares
        if not (is_ul_config or is_uphase_config):
            # Type 0
            return 0, self.topological_corners

        # Find the last topological corner
        topological_non_geometric_corner = (
            set(self.topological_corners) - set(topological_geometric_corners)
        ).pop()

        # Based on the orientation of the block, deduce the long and short edge indices
        long_edge_idx, short_edge_idx = (0, 1) if is_block_horizontal else (1, 0)

        # Find two topological-geometric corners that reside on the same long edge
        long_edge_geometric_corners = next(
            (corner_1, corner_2)
            for (corner_1, corner_2) in combinations(topological_geometric_corners, 2)
            if corner_1[short_edge_idx] == corner_2[short_edge_idx]
        )

        # Find the third topological-geometric corner
        short_edge_geometric_corner = next(
            corner
            for corner in topological_geometric_corners
            if corner not in long_edge_geometric_corners
        )

        # Sort the long_edge_geometric_corners so that the first returned corner is
        # the long_end.
        long_edge_geometric_corners = sorted(
            long_edge_geometric_corners,
            reverse=topological_non_geometric_corner[long_edge_idx]
            > short_edge_geometric_corner[long_edge_idx],
        )

        # Verify that the distances (along the long edge) between the last topological
        # corner and two long_edge_corners differ by no larger than one unit.
        # If it's not, the config is not U or L
        middle_long_dist = abs(
            topological_non_geometric_corner[long_edge_idx]
            - long_edge_geometric_corners[0][long_edge_idx]
        )  # distance from middle_edge corner to long_end corner
        middle_angle_dist = abs(
            long_edge_geometric_corners[1][long_edge_idx]
            - topological_non_geometric_corner[long_edge_idx]
        )  # distance from angle_corner to middle_edge corner
        dist_diff = (
            middle_angle_dist - middle_long_dist
        )  # no abs because only one case is supported
        if dist_diff not in [0, 1]:
            # Type 0
            return 0, self.topological_corners

        # If the short edge index of the topological non-geometric corner matches:
        # - the short edge geometric corner, then it is a U shape
        # - one of the long edge geometric corners, then it is a L shape
        if (
            topological_non_geometric_corner[short_edge_idx]
            == short_edge_geometric_corner[short_edge_idx]
        ) and dist_diff == 0:
            # Type 2 (U shape)
            config = 2
        elif (
            topological_non_geometric_corner[short_edge_idx]
            == long_edge_geometric_corners[0][short_edge_idx]
        ) and dist_diff == 0:
            # Type 3 (L shape)
            config = 3
        elif (
            topological_non_geometric_corner[short_edge_idx]
            == short_edge_geometric_corner[short_edge_idx]
        ) and dist_diff == 1:
            # Type 4 (U shape phase gate)
            config = 4
        else:
            raise RuntimeError("Something went wrong with the configuration detection.")

        # Put all the corners together
        corners = (
            long_edge_geometric_corners[0],  # long_end
            topological_non_geometric_corner,  # middle_edge
            long_edge_geometric_corners[1],  # angle
            short_edge_geometric_corner,  # short_edge
        )

        return config, corners

    def get_shifted_equivalent_logical_operator(
        self, initial_operator: PauliOperator, new_upleft_qubit: tuple[int, ...]
    ) -> tuple[PauliOperator, tuple[Stabilizer, ...]]:
        """
        Shifts the initial operator to a valid operator in the block that contains
        new_upleft_qubit.
        NOTE: this function currently assumes that the block is a square rotated surface
        code block with a single logical X and Z operator.

        Parameters
        ----------
        initial_operator : PauliOperator
            Initial logical operator for which the equivalent logical operator in the
            block should be found.
        new_upleft_qubit : tuple[int, ...]
            New top left qubit of the operator in the block.

        Returns
        -------
        tuple[PauliOperator, tuple[Stabilizer, ...]]
            Equivalent logical operator in the block and the stabilizers that are
            required to go from the initial to the new operator.

        Raises
        ------
        NotImplementedError
            If the block has more than one logical X or Z operator.
        ValueError
            If the new upleft qubit is not part of the data qubits of the block.
        ValueError
            If the initial operator is not one of the logical operators of the block.
        ValueError
            If the new upleft qubit is not part of the correct boundary.
        ValueError
            If the shift vector has more than a single non-zero dimension (ill defined
            logical operators to begin with).
        """
        if len(self.logical_x_operators) != 1 or len(self.logical_z_operators) != 1:
            raise NotImplementedError(
                "This function currently only supports blocks with a single logical "
                "X and Z operator."
            )

        if new_upleft_qubit not in self.data_qubits:
            raise ValueError(
                f"The new upleft qubit {new_upleft_qubit} is not part of the data "
                "qubits of the block."
            )

        if initial_operator not in [
            self.logical_x_operators[0],
            self.logical_z_operators[0],
        ]:
            raise ValueError(
                "The initial operator must be one of the logical operators of the "
                "block."
            )

        # We assume all Xs or all Zs pauli operators
        # The new upper left qubit needs to be part of the right boundary, Y is valid.
        # E.g. Z logical has an X or Y at it upper left most qubit.
        if self.pauli_charges[new_upleft_qubit] == initial_operator.pauli[0]:
            raise ValueError(
                f"The new upleft qubit {new_upleft_qubit} is not part of the correct "
                "boundary."
            )

        initial_upleft_qubit = min(
            initial_operator.data_qubits, key=lambda x: x[0] + x[1]
        )
        shift_vector = tuple(
            coord1 - coord2
            for coord1, coord2 in zip(
                new_upleft_qubit, initial_upleft_qubit, strict=True
            )
        )

        new_logical_operator = PauliOperator(
            pauli=initial_operator.pauli,
            data_qubits=[
                tuple(
                    coord1 + shift
                    for coord1, shift in zip(qb, shift_vector, strict=True)
                )
                for qb in initial_operator.data_qubits
            ],
        )

        # Find all stabilizers in the block that are included between the initial and
        # new operator
        def vector_range(vector: tuple[int, ...]) -> list[tuple[int, ...]]:
            """
            Generate vectors from (0, ..., 0) to (0, ..., k, ..., 0) where k is the only
            non-zero element in the input vector. The position of k in the tuple may
            vary. If k is negative, the range goes from k to 0.
            E.g.:
            vector_range((0, 5)) -> [(0, 0), (0, 1), (0, 2), (0, 3), (0, 4), (0, 5)]
            vector_range((-3, 0, 0)) -> [(-3, 0, 0), (-2, 0, 0), (-1, 0, 0), (0, 0, 0)]
            """
            # Find the position and value of the non-zero element in the vector
            k_position = next(i for i, x in enumerate(vector) if x != 0)
            k = vector[k_position]

            # Determine the range based on the sign of k
            if k > 0:
                range_values = range(k + 1)
            else:
                range_values = range(k, 1)

            return [
                tuple(i if j == k_position else 0 for j in range(len(vector)))
                for i in range_values
            ]

        # Test that shift vector has at most a single non-zero dimension
        n_non_zero_dims = sum(1 for x in shift_vector if x != 0)
        match n_non_zero_dims:
            # The two qubits have the same coordinates
            case 0:
                return new_logical_operator, ()
            # The two qubits have a single non-zero dimension, they are aligned
            case 1:
                qubits_in_between = [
                    tuple(
                        qubit_coord + shift_coord
                        for (qubit_coord, shift_coord) in zip(q, v, strict=True)
                    )
                    for q in initial_operator.data_qubits
                    for v in vector_range(shift_vector)
                ]
                # Here we assume all Xs or all Zs stabilizers
                stabilizers_in_between = [
                    stab
                    for stab in self.stabilizers
                    if all(q in qubits_in_between for q in stab.data_qubits)
                    and stab.pauli[0] == initial_operator.pauli[0]
                ]
                return new_logical_operator, tuple(stabilizers_in_between)
            # The two qubits have more than a single non-zero dimension,
            # they are not aligned.
            case _:
                raise ValueError(
                    "The shift vector must have at most a single non-zero dimension."
                )

    def rename(self, name: str) -> RotatedSurfaceCode:
        return super().rename(name)

    def shift(
        self, position: tuple[int, ...], new_label: str | None = None
    ) -> RotatedSurfaceCode:
        return super().shift(position, new_label)

    def __eq__(self, other) -> bool:  # pylint: disable=useless-parent-delegation
        return super().__eq__(other)

    @cached_property
    def stabilizers_labels(self) -> dict[str, dict[str, tuple[int, ...]]]:
        return super().stabilizers_labels
